"""Validators for checking if data model is AI-ready."""

from cognite.neat._data_model.models.dms._data_types import EnumProperty
from cognite.neat._data_model.rules.dms._base import DataModelRule
from cognite.neat._issues import Recommendation

BASE_CODE = "NEAT-DMS-AI-READINESS"


class DataModelMissingName(DataModelRule):
    """Validates that data model has a human-readable name.

    ## What it does
    Validates that the data model has a human-readable name.

    ## Why is this bad?
    Often the data model ids are technical identifiers, abbreviations, etc.
    A missing name makes it harder for users (humans or machines) to understand what the data model represents.
    Providing a clear name improves usability, maintainability, searchability, and AI-readiness.

    ## Example
    A data model has an id IEC61400-25-2 but no name. Users may find it difficult to understand what this data model
    represents. However adding a name "Wind Energy Information Model" would increase clarity and usability.
    """

    code = f"{BASE_CODE}-001"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        if not self.validation_resources.merged_data_model.name:
            recommendations.append(
                Recommendation(
                    message="Data model is missing a human-readable name.",
                    fix="Add a clear and concise name to the data model.",
                    code=self.code,
                )
            )

        return recommendations


class DataModelMissingDescription(DataModelRule):
    """Validates that data model has a human-readable description.

    ## What it does
    Validates that the data model has a human-readable description.

    ## Why is this bad?
    A missing description makes it harder for users (humans or machines) to understand the purpose and scope
    of the data model. The description provides important context about what domain the data model covers,
    what use cases it supports, and how it should be used.

    ## Example
    A data model has an id CIM, with name Common Information Model, but no description. Users may find it difficult to
    understand what this data model represents, unless extra context is provided. In this particualar case, name
    does not provide sufficient information, as it is too generic, that this data model is focused on the
    electrical power systems domain. However, providing a description such as:
    "The Common Information Model (CIM) is a standard developed by IEC for representing power system
    components and their relationships. It is widely used in the electrical utility industry for data
    exchange and system modeling." would greatly improve clarity and usability.
    """

    code = f"{BASE_CODE}-002"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        if not self.validation_resources.merged_data_model.description:
            recommendations.append(
                Recommendation(
                    message="Data model is missing a description.",
                    fix="Add a clear and concise description to the data model.",
                    code=self.code,
                )
            )

        return recommendations


class ViewMissingName(DataModelRule):
    """Validates that a View has a human-readable name.

    ## What it does
    Validates that each view in the data model has a human-readable name.

    ## Why is this bad?
    A missing name makes it harder for users (humans or machines) to understand the purpose of the view.
    This is important as views' external ids are often based on technical identifiers, abbreviations, etc.
    Providing a clear name improves usability, maintainability, searchability, and AI-readiness.

    ## Example
    A view has an id CFIHOS-30000038 but no name. Users may find it difficult to understand what this view represents,
    unless they look up the id in documentation or other resources. Adding name "Pump" would increase clarity and
    usability.
    """

    code = f"{BASE_CODE}-003"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for view_ref in self.validation_resources.merged_data_model.views or []:
            view = self.validation_resources.select_view(view_ref)

            if view is None:
                raise RuntimeError(f"{type(self).__name__}: View {view_ref!s} not found. This is a bug.")

            if not view.name:
                recommendations.append(
                    Recommendation(
                        message=f"View {view_ref!s} is missing a human-readable name.",
                        fix="Add a clear and concise name to the view.",
                        code=self.code,
                    )
                )

        return recommendations


class ViewMissingDescription(DataModelRule):
    """Validates that a View has a human-readable description.

    ## What it does
    Validates that each view in the data model has a human-readable description.

    ## Why is this bad?
    A missing description makes it harder for users (humans or machines) to understand in what context the view
    should be used. The description can provide important information about the view's purpose, scope, and usage.


    ## Example
    A view Site has no description. Users may find it difficult to understand what this view represents, unless
    extra context is provided. Even if we know that Site is used in the context of wind energy developments, a
    description is necessary as it can be used in various context within the same domain such as:

    Option 1 — Project area
    This view represents a geographical area where wind energy projects are developed and managed.

    Option 2 — Lease area
    The legally defined lease area allocated for offshore wind development.

    Option 3 — Measurement site
    A specific location where wind measurements (e.g., LiDAR, met mast) are collected.

    """

    code = f"{BASE_CODE}-004"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for view_ref in self.validation_resources.merged_data_model.views or []:
            view = self.validation_resources.select_view(view_ref)

            if view is None:
                raise RuntimeError(f"{type(self).__name__}: View {view_ref!s} not found. This is a bug.")

            if not view.description:
                recommendations.append(
                    Recommendation(
                        message=f"View {view_ref!s} is missing a description.",
                        fix="Add a clear and concise description to the view.",
                        code=self.code,
                    )
                )

        return recommendations


class ViewPropertyMissingName(DataModelRule):
    """Validates that a view property has a human-readable name.

    ## What it does
    Validates that each view property in the data model has a human-readable name.

    ## Why is this bad?
    A missing name makes it harder for users (humans or machines) to understand the purpose of the view property.
    This is important as view property's ids are often based on technical identifiers, abbreviations, etc.
    Providing a clear name improves usability, maintainability, searchability, and AI-readiness.

    ## Example
    A view WindTurbine has a property pc which has no name. Users may find it difficult to understand what this view
    property represents, unless they look up the id in documentation or other resources. Adding name "power curve"
    would increase clarity and usability.
    """

    code = f"{BASE_CODE}-005"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for view_ref in self.validation_resources.merged_data_model.views or []:
            view = self.validation_resources.select_view(view_ref)

            if view is None:
                raise RuntimeError(f"{type(self).__name__}: View {view_ref!s} not found. This is a bug.")

            if not view.properties:
                continue

            for prop_ref, definition in view.properties.items():
                if not definition.name:
                    recommendations.append(
                        Recommendation(
                            message=f"View {view_ref!s} property {prop_ref!s} is missing a human-readable name.",
                            fix="Add a clear and concise name to the view property.",
                            code=self.code,
                        )
                    )

        return recommendations


class ViewPropertyMissingDescription(DataModelRule):
    """Validates that a View property has a human-readable description.

    ## What it does
    Validates that each view property in the data model has a human-readable description.

    ## Why is this bad?
    A missing description makes it harder for users (humans or machines) to understand in what context the view property
    should be used. The description can provide important information about the view property's purpose,
    scope, and usage.


    ## Example
    A view WindTurbine has a property status with no description. Users may find it difficult to understand what this
    property represents, unless extra context is provided. Even if we know that status is related to wind turbine
    operations, a description is necessary as it can have different meanings in various contexts:

    Option 1 — Operational status
    Current operational state of the wind turbine (e.g., running, stopped, maintenance, fault).

    Option 2 — Connection status
    Grid connection status indicating whether the turbine is connected to the electrical grid.

    Option 3 — Availability status
    Availability state for production indicating whether the turbine is available for power generation.

    """

    code = f"{BASE_CODE}-006"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for view_ref in self.validation_resources.merged_data_model.views or []:
            view = self.validation_resources.select_view(view_ref)

            if view is None:
                raise RuntimeError(f"{type(self).__name__}: View {view_ref!s} not found. This is a bug.")

            if not view.properties:
                continue

            for prop_ref, definition in view.properties.items():
                if not definition.description:
                    recommendations.append(
                        Recommendation(
                            message=f"View {view_ref!s} property {prop_ref!s} is missing a description.",
                            fix="Add a clear and concise description to the view property.",
                            code=self.code,
                        )
                    )

        return recommendations


class EnumerationMissingName(DataModelRule):
    """Validates that an enumeration has a human-readable name.

    ## What it does
    Validates that each enumeration value in the data model has a human-readable name.

    ## Why is this bad?
    A missing name makes it harder for users (humans or machines) to understand the purpose of the enumeration value.
    This is important as enumeration values are often technical codes or abbreviations, and a clear name improves
    usability, maintainability, searchability, and AI-readiness.

    ## Example
    An enumeration value with id "NOM" in a wind turbine operational mode property has no name. Users may find it
    difficult to understand what this value represents. Adding name "Normal Operation" would increase clarity
    and usability.
    """

    code = f"{BASE_CODE}-007"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for container_ref in self.validation_resources.merged.containers:
            container = self.validation_resources.select_container(container_ref)

            if not container:
                raise RuntimeError(f"{type(self).__name__}: Container {container_ref!s} not found. This is a bug.")

            for prop_ref, definition in container.properties.items():
                if not isinstance(definition.type, EnumProperty):
                    continue

                for value, enum_def in definition.type.values.items():
                    if not enum_def.name:
                        recommendations.append(
                            Recommendation(
                                message=(
                                    f"Enumeration value {value!r} in property {prop_ref!s} of container "
                                    f"{container_ref!s} is missing a human-readable name."
                                ),
                                fix="Add a clear and concise name to the enumeration value.",
                                code=self.code,
                            )
                        )

        return recommendations


class EnumerationMissingDescription(DataModelRule):
    """Validates that an enumeration value has a human-readable description.

    ## What it does
    Validates that each enumeration value in the data model has a human-readable description.

    ## Why is this bad?
    A missing description makes it harder for users (humans or machines) to understand the meaning and context
    of the enumeration value. The description can provide important information about when and how the value
    should be used, especially when enumeration values are technical codes or abbreviations.

    ## Example
    An enumeration value "NOM" in a wind turbine operational mode property has no description. Users may find it
    difficult to understand what this value represents without additional context. Even with a name like
    "Normal Operation", the description is valuable as it can clarify specifics:

    Option 1 — Basic definition
    The turbine is operating normally and generating power according to its power curve.

    Option 2 — Detailed operational context
    The turbine is in normal operation mode, actively generating power with all systems functioning within
    specified parameters and connected to the grid.

    Option 3 — Contrasting with other modes
    Standard operating mode where the turbine follows the power curve and responds to grid commands,
    as opposed to maintenance mode or fault conditions.
    """

    code = f"{BASE_CODE}-008"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for container_ref in self.validation_resources.merged.containers:
            container = self.validation_resources.select_container(container_ref)
            if not container:
                raise RuntimeError(f"{self.__class__.__name__}: Container {container_ref!s} not found. This is a bug.")

            for prop_ref, definition in container.properties.items():
                if not isinstance(definition.type, EnumProperty):
                    continue

                for value, enum_def in definition.type.values.items():
                    if not enum_def.description:
                        recommendations.append(
                            Recommendation(
                                message=(
                                    f"Enumeration value {value!r} in property {prop_ref!s} of container "
                                    f"{container_ref!s} is missing a human-readable description."
                                ),
                                fix="Add a clear and concise description to the enumeration value.",
                                code=self.code,
                            )
                        )

        return recommendations



import sys
from datetime import datetime, timezone
from typing import Any

from pydantic import BaseModel, Field, field_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from cognite.neat._client import NeatClient
from cognite.neat._data_model.models.dms import (
    ContainerReference,
    ContainerRequest,
    DataModelReference,
    DataModelRequest,
    NodeReference,
    RequestSchema,
    SpaceReference,
    SpaceRequest,
    ViewReference,
    ViewRequest,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class SchemaSnapshot(BaseModel, extra="ignore"):
    timestamp: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    data_model: dict[DataModelReference, DataModelRequest] = Field(default_factory=dict)
    views: dict[ViewReference, ViewRequest] = Field(default_factory=dict)
    containers: dict[ContainerReference, ContainerRequest] = Field(default_factory=dict)
    spaces: dict[SpaceReference, SpaceRequest] = Field(default_factory=dict)
    node_types: dict[NodeReference, NodeReference] = Field(default_factory=dict)

    @field_serializer("data_model", "views", "containers", "spaces", "node_types", mode="plain")
    @classmethod
    def make_hashable_keys(cls, value: dict, info: FieldSerializationInfo) -> dict[str, Any]:
        output: dict[str, Any] = {}
        for key, val in value.items():
            dumped_value = val.model_dump(**vars(info))
            output[str(key)] = dumped_value
        return output

    def merge(self, cdf: Self) -> Self:
        """Merge another SchemaSnapshot into this one, prioritizing this snapshot's data."""
        merged = self.model_copy(deep=True)

        for model_ref, local_model in merged.data_model.items():
            if model_ref not in cdf.data_model:
                continue
            cdf_model = cdf.data_model[model_ref]
            if new_views := (set(cdf_model.views or []) - set(local_model.views or [])):
                for view_ref in new_views:
                    if cdf_view := cdf.views.get(view_ref):
                        merged.views[view_ref] = cdf_view
            # We append the local views at the end of the CDF views.
            local_model.views = list(dict.fromkeys((cdf_model.views or []) + (local_model.views or [])).keys())

        # Update local views with additional properties and implements from CDF views
        for view_ref, view in merged.views.items():
            if cdf_view := cdf.views.get(view_ref):
                if cdf_only_containers := cdf_view.used_containers - set(view.used_containers):
                    for cdf_only_container_ref in cdf_only_containers:
                        if (
                            cdf_container := cdf.containers.get(cdf_only_container_ref)
                        ) and cdf_only_container_ref not in merged.containers:
                            merged.containers[cdf_only_container_ref] = cdf_container

                view.properties = {**cdf_view.properties, **view.properties}

                # update implements
                if cdf_view.implements:
                    view.implements = list(dict.fromkeys(cdf_view.implements + (view.implements or [])).keys())

        for container_ref, container in merged.containers.items():
            if cdf_container := cdf.containers.get(container_ref):
                container.properties = {**cdf_container.properties, **container.properties}
                container.constraints = {**(cdf_container.constraints or {}), **(container.constraints or {})} or None

        return merged

    @classmethod
    def fetch_cdf_data_model(cls, client: NeatClient, data_model: RequestSchema) -> Self:
        """Fetch the latest data model, views, containers, and spaces from CDF based on the provided RequestSchema."""
        now = datetime.now(timezone.utc)
        space_ids = [space.as_reference() for space in data_model.spaces]
        cdf_spaces = client.spaces.retrieve(space_ids)

        container_refs = [c.as_reference() for c in data_model.containers]
        cdf_containers = client.containers.retrieve(container_refs)

        view_refs = [v.as_reference() for v in data_model.views]
        cdf_views = client.views.retrieve(view_refs)

        dm_ref = data_model.data_model.as_reference()
        cdf_data_models = client.data_models.retrieve([dm_ref])

        nodes = [node_type for view in cdf_views for node_type in view.node_types]
        return cls(
            timestamp=now,
            data_model={dm.as_reference(): dm.as_request() for dm in cdf_data_models},
            views={view.as_reference(): view.as_request() for view in cdf_views},
            containers={container.as_reference(): container.as_request() for container in cdf_containers},
            spaces={space.as_reference(): space.as_request() for space in cdf_spaces},
            node_types={node: node for node in nodes},
        )

    @classmethod
    def fetch_entire_cdf(cls, client: NeatClient) -> Self:
        """Fetch the entire data model, views, containers, and spaces from CDF."""
        now = datetime.now(timezone.utc)
        all_views = client.views.list(
            all_versions=True, include_global=True, include_inherited_properties=False, limit=None
        )
        nodes = [node_type for view in all_views for node_type in view.node_types]
        return cls(
            # TODO: spaces and data_models should be update after updating list methods for unlimited no
            spaces={
                response.as_reference(): response.as_request()
                for response in client.spaces.list(include_global=True, limit=1000)
            },
            data_model={
                response.as_reference(): response.as_request()
                for response in client.data_models.list(all_versions=True, include_global=True, limit=1000)
            },
            views={response.as_reference(): response.as_request() for response in all_views},
            containers={
                response.as_reference(): response.as_request()
                for response in client.containers.list(include_global=True, limit=None)
            },
            node_types={node_type: node_type for node_type in nodes},
            timestamp=now,
        )



import math
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from itertools import chain, combinations
from typing import Literal, TypeAlias, TypeVar

import networkx as nx

from cognite.neat._client.data_classes import SpaceStatisticsResponse
from cognite.neat._data_model._constants import COGNITE_SPACES
from cognite.neat._data_model._snapshot import SchemaSnapshot
from cognite.neat._data_model.models.dms._constraints import RequiresConstraintDefinition
from cognite.neat._data_model.models.dms._container import ContainerPropertyDefinition, ContainerRequest
from cognite.neat._data_model.models.dms._data_types import DirectNodeRelation
from cognite.neat._data_model.models.dms._limits import SchemaLimits
from cognite.neat._data_model.models.dms._references import (
    ContainerDirectReference,
    ContainerReference,
    ViewDirectReference,
    ViewReference,
)
from cognite.neat._data_model.models.dms._view_property import (
    EdgeProperty,
    ReverseDirectRelationProperty,
    ViewCorePropertyRequest,
    ViewRequestProperty,
)
from cognite.neat._data_model.models.dms._views import ViewRequest
from cognite.neat._utils.useful_types import ModusOperandi, T_Reference

# Type aliases for better readability
ViewsByReference: TypeAlias = dict[ViewReference, ViewRequest]
ContainersByReference: TypeAlias = dict[ContainerReference, ContainerRequest]
AncestorsByReference: TypeAlias = dict[ViewReference, set[ViewReference]]

ReverseToDirectMapping: TypeAlias = dict[
    tuple[ViewReference, str], tuple[ViewReference, ContainerDirectReference | ViewDirectReference]
]
ConnectionEndNodeTypes: TypeAlias = dict[tuple[ViewReference, str], ViewReference | None]


ResourceSource = Literal["auto", "merged", "cdf", "both"]

_NodeT = TypeVar("_NodeT", ContainerReference, ViewReference)


class RequiresChangeStatus(Enum):
    """Status of requires constraint changes for a view."""

    OPTIMAL = "optimal"  # Already optimized, no changes needed
    CHANGES_AVAILABLE = "changes_available"  # Recommendations available
    UNSOLVABLE = "unsolvable"  # Structural issue - can't create connected hierarchy
    NO_MODIFIABLE_CONTAINERS = "no_modifiable_containers"  # All containers are immutable


@dataclass
class ResolvedReverseDirectRelation:
    """Resolved context for a reverse direct relation, including container-level information."""

    reverse_view_ref: ViewReference
    reverse_property_id: str
    direct_view_ref: ViewReference
    through_property_id: str
    direct_property: ViewCorePropertyRequest
    container_ref: ContainerReference
    container_property_id: str
    container: ContainerRequest | None
    container_property: ContainerPropertyDefinition | None


@dataclass
class RequiresChangesForView:
    """Result of computing requires constraint changes for a view."""

    to_add: set[tuple[ContainerReference, ContainerReference]]
    to_remove: set[tuple[ContainerReference, ContainerReference]]
    status: RequiresChangeStatus


class ValidationResources:
    def __init__(
        self,
        modus_operandi: ModusOperandi,
        local: SchemaSnapshot,
        cdf: SchemaSnapshot,
        limits: SchemaLimits | None = None,
        space_statistics: SpaceStatisticsResponse | None = None,
    ) -> None:
        self._modus_operandi = modus_operandi
        self.limits = limits or SchemaLimits()
        self.space_statistics = space_statistics

        self.local = local
        self.cdf = cdf

        if self._modus_operandi == "additive":
            self.merged = self.local.merge(self.cdf)
        elif self._modus_operandi == "rebuild":
            self.merged = local.model_copy(deep=True)
        else:
            raise RuntimeError(f"ValidationResources: Unknown modus_operandi: {self._modus_operandi}. This is a bug!")

        # need this shortcut for easier access and also to avoid mypy to complains
        self.merged_data_model = self.merged.data_model[next(iter(self.merged.data_model.keys()))]

        # For caching of expanded views
        self._expanded_views_cache: dict[ViewReference, ViewRequest] = {}

    def select_view(
        self, view_ref: ViewReference, property_: str | None = None, source: ResourceSource = "auto"
    ) -> ViewRequest | None:
        """Select view definition based on source strategy and optionally filter by property.

        Selection prioritize merged view over CDF if both are available, as merged view represents the effective
        definition which will be a result when the local schema is deployed to CDF.


        Args:
            view_ref: The view to select
            property_: Optional property name to filter views that contain this property
            source: The source strategy to use, options are: "auto", "local", "cdf", "both", where "auto" means
                that the selection is driven by the modus_operandi of the validation resources.

        Returns:
            The selected ViewRequest or None if not found.
        """

        check_merged, check_cdf = self._resolve_resource_sources(view_ref, source)

        merged_view = self.merged.views.get(view_ref) if check_merged else None
        cdf_view = self.cdf.views.get(view_ref) if check_cdf else None

        if property_ is None:
            return merged_view or cdf_view

        # Filtering based on the property presence
        # Try views with the property first, then any available view where merged view is prioritized
        candidates = chain(
            (v for v in (merged_view, cdf_view) if v and v.properties and property_ in v.properties),
            (v for v in (merged_view, cdf_view) if v),
        )

        return next(candidates, None)

    def select_container(
        self, container_ref: ContainerReference, property_: str | None = None, source: ResourceSource = "auto"
    ) -> ContainerRequest | None:
        """Select container definition based on source strategy and optionally filter by property.

        Selection prioritize merged container over CDF if both are available, as merged container represents
        the effective definition which will be a result when the local schema is deployed to CDF.

        Args:
            container_ref: The container to select
            property_: Optional property name to filter containers that contain this property
            source: The source strategy to use, options are: "auto", "local", "cdf", "both", where "auto" means
                that the selection is driven by the modus_operandi of the validation resources.

        Returns:
            The selected ContainerRequest or None if not found.
        """

        check_merged, check_cdf = self._resolve_resource_sources(container_ref, source)

        merged_container = self.merged.containers.get(container_ref) if check_merged else None
        cdf_container = self.cdf.containers.get(container_ref) if check_cdf else None

        if property_ is None:
            return merged_container or cdf_container

        # Try containers with the property first, then any available container
        candidates = chain(
            (c for c in (merged_container, cdf_container) if c and c.properties and property_ in c.properties),
            (c for c in (merged_container, cdf_container) if c),
        )

        return next(candidates, None)

    def _resolve_resource_sources(
        self, resource_ref: ViewReference | ContainerReference, source: ResourceSource
    ) -> tuple[bool, bool]:
        """
        Determine which resource sources (merged and/or CDF) to check based on the source parameter.

        Args:
            resource_ref: The resource reference to check (ViewReference or ContainerReference)
            source: The source strategy to use

        Returns:
            Tuple of (check_merged, check_cdf) booleans indicating which sources to check
        """
        if source == "auto":
            # Auto mode: driven by data modeling modus (approach)
            # If elements is in the schema space, we check merged, else we check CDF

            in_schema_space = resource_ref.space == self.merged_data_model.space

            if self._modus_operandi == "additive":
                # In additive modus, schema space means local additions on top of CDF
                # always check CDF, while do not check merged if resource is not in schema space
                check_merged = in_schema_space
                check_cdf = True
            elif self._modus_operandi == "rebuild":
                # In rebuild modus, schema space means the full desired state is in local schema (i.e., merged)
                # you are not adding to CDF, but replacing it, so never check CDF for schema space resources
                check_merged = in_schema_space
                check_cdf = not in_schema_space
            else:
                raise RuntimeError(
                    f"_resolve_resource_sources: Unknown modus_operandi: {self._modus_operandi}. This is a bug!"
                )

        elif source == "merged":
            check_merged = True
            check_cdf = False
        elif source == "cdf":
            check_merged = False
            check_cdf = True
        elif source == "both":
            check_merged = True
            check_cdf = True
        else:
            raise RuntimeError(f"_resolve_resource_sources: Unknown source: {source}. This is a bug!")

        return check_merged, check_cdf

    @cached_property
    def ancestors_by_view(self) -> dict[ViewReference, list[ViewReference]]:
        """
        Create a mapping of each view to its list of ancestors.

        Returns:
            Dictionary mapping each ViewReference to its list of ancestor ViewReferences
        """
        ancestors_mapping: dict[ViewReference, list[ViewReference]] = {}

        if not self.merged_data_model.views:
            return ancestors_mapping

        for view in self.merged_data_model.views:
            ancestors_mapping[view] = self.view_ancestors(view)
        return ancestors_mapping

    def view_ancestors(
        self, offspring: ViewReference, ancestors: list[ViewReference] | None = None, source: ResourceSource = "auto"
    ) -> list[ViewReference]:
        """
        Recursively find all ancestors of a given view by traversing the implements hierarchy.
        Handles branching to explore all possible ancestor paths.

        Args:
            offspring: The view to find ancestors for
            ancestors: Accumulated list of ancestors (used internally for recursion)

        Returns:
            List of all ancestor ViewReferences
        """
        if ancestors is None:
            ancestors = []

        # Determine which view definition to use based on space and modus operandi
        view_definition = self.select_view(view_ref=offspring, source=source)

        # Base case: no view definition or no implements
        if not view_definition or not view_definition.implements:
            return ancestors

        # Explore all parent branches
        for parent in view_definition.implements:
            if parent not in ancestors:
                ancestors.append(parent)
                # Recursively explore this branch
                self.view_ancestors(parent, ancestors)

        return ancestors

    def is_ancestor(self, offspring: ViewReference, ancestor: ViewReference) -> bool:
        return ancestor in self.view_ancestors(offspring)

    def _expand_view(self, view_ref: ViewReference) -> ViewRequest | None:
        """Expand a view by including properties from its ancestors.

        Args:
            view_ref: The view to expand.

        Returns:
            ViewRequest with expanded properties, or None if view not found.
        """
        view = self.select_view(view_ref)

        if not view:
            return None

        # Create a deep copy to avoid mutating the original
        expanded_view = view.model_copy(deep=True)

        # Get all ancestor properties (oldest to newest)
        ancestor_refs = self.view_ancestors(view_ref)
        ancestor_properties: dict[str, ViewRequestProperty] = {}

        # Collect properties from ancestors, overriding with newer ancestors properties
        for ancestor_ref in reversed(ancestor_refs):
            ancestor = self.select_view(ancestor_ref)
            if ancestor and ancestor.properties:
                ancestor_properties.update(ancestor.properties)

        # Merge: ancestor properties first, then override with view's own properties
        if ancestor_properties:
            if not expanded_view.properties:
                expanded_view.properties = {}

            # Ancestor properties are base, view properties override
            expanded_view.properties = {**ancestor_properties, **expanded_view.properties}

        return expanded_view

    def expand_view_properties(self, view_ref: ViewReference) -> ViewRequest | None:
        """Get a mapping of view references to their corresponding properties, both directly defined and inherited
        from ancestor views through implements."""

        if view_ref not in self._expanded_views_cache:
            expanded_view = self._expand_view(view_ref)
            if expanded_view:
                self._expanded_views_cache[view_ref] = expanded_view

        return self._expanded_views_cache.get(view_ref)

    @cached_property
    def referenced_containers(self) -> set[ContainerReference]:
        """Get a set of all container references used by the views in the local data model."""

        referenced_containers: set[ContainerReference] = set()

        if not self.merged_data_model.views:
            return referenced_containers

        for view_ref in self.merged_data_model.views:
            view = self.select_view(view_ref)
            # This should never happen, if it happens, it's a bug
            if not view:
                raise RuntimeError(f"referenced_containers: View {view_ref!s} not found. This is a bug!")

            if not view.properties:
                continue
            for property_ in view.properties.values():
                if isinstance(property_, ViewCorePropertyRequest):
                    referenced_containers.add(property_.container)

        return referenced_containers

    @cached_property
    def reverse_to_direct_mapping(
        self,
    ) -> dict[tuple[ViewReference, str], tuple[ViewReference, ContainerDirectReference | ViewDirectReference]]:
        """Get a mapping of reverse direct relations to their corresponding source view and 'through' property."""

        bidirectional_connections: dict[
            tuple[ViewReference, str], tuple[ViewReference, ContainerDirectReference | ViewDirectReference]
        ] = {}

        if self.merged_data_model.views:
            for view_ref in self.merged_data_model.views:
                view = self.select_view(view_ref)

                # This should never happen, if it happens, it's a bug
                if not view:
                    raise RuntimeError(f"reverse_to_direct_mapping: View {view_ref!s} not found. This is a bug!")

                if not view.properties:
                    continue
                for prop_ref, property_ in view.properties.items():
                    # reverse direct relation
                    if isinstance(property_, ReverseDirectRelationProperty):
                        bidirectional_connections[(view_ref, prop_ref)] = (
                            property_.source,
                            property_.through,
                        )

        return bidirectional_connections

    @staticmethod
    def normalize_through_reference(
        source_view_ref: ViewReference, through: ContainerDirectReference | ViewDirectReference
    ) -> ViewDirectReference:
        """Normalize through reference to ViewDirectReference for consistent processing.

        When a reverse direct relation uses a ContainerDirectReference, we convert it to
        a ViewDirectReference using the source view. This enables consistent handling
        in validators regardless of how the 'through' was originally specified.
        """
        if isinstance(through, ContainerDirectReference):
            return ViewDirectReference(source=source_view_ref, identifier=through.identifier)
        return through

    @cached_property
    def resolved_reverse_direct_relations(self) -> list[ResolvedReverseDirectRelation]:
        """Get all reverse direct relations with their resolved context.

        This property traverses from reverse direct relation → source view → container,
        resolving all references along the way.

        Includes reverse direct relations where view-level resolution succeeded:
        - The source view exists and can be expanded
        - The through property exists in the source view
        - The through property is a ViewCorePropertyRequest (direct relation)

        Container-level fields may be None if resolution failed at that level.
        """
        result: list[ResolvedReverseDirectRelation] = []

        for (target_view_ref, reverse_prop_id), (
            source_view_ref,
            through,
        ) in self.reverse_to_direct_mapping.items():
            through_normalized = self.normalize_through_reference(source_view_ref, through)

            # Get expanded source view to include inherited properties
            source_view_expanded = self.expand_view_properties(source_view_ref)
            if not source_view_expanded or not source_view_expanded.properties:
                continue

            if through_normalized.identifier not in source_view_expanded.properties:
                continue

            source_property = source_view_expanded.properties[through_normalized.identifier]

            # Must be a core property (direct relation)
            if not isinstance(source_property, ViewCorePropertyRequest):
                continue

            container_ref = source_property.container
            container_property_id = source_property.container_property_identifier

            # Resolve container - may be None if missing
            container = self.select_container(container_ref, container_property_id)

            # Resolve container property - may be None if container missing or property missing
            container_property = None
            if container and container.properties:
                container_property = container.properties.get(container_property_id)

            result.append(
                ResolvedReverseDirectRelation(
                    reverse_view_ref=target_view_ref,
                    reverse_property_id=reverse_prop_id,
                    direct_view_ref=source_view_ref,
                    through_property_id=through_normalized.identifier,
                    direct_property=source_property,
                    container_ref=container_ref,
                    container_property_id=container_property_id,
                    container=container,
                    container_property=container_property,
                )
            )

        return result

    @property
    def connection_end_node_types(self) -> dict[tuple[ViewReference, str], ViewReference | None]:
        """Get a mapping of view references to their corresponding ViewRequest objects."""

        connection_end_node_types: dict[tuple[ViewReference, str], ViewReference | None] = {}

        if self.merged_data_model.views:
            for view_ref in self.merged_data_model.views:
                view = self.select_view(view_ref)
                if not view:
                    raise RuntimeError(f"View {view_ref!s} not found. This is a bug!")

                if not view.properties:
                    continue

                for prop_ref, property_ in view.properties.items():
                    # direct relation
                    if isinstance(property_, ViewCorePropertyRequest):
                        # explicit set of end node type via 'source' which is View reference
                        if property_.source:
                            connection_end_node_types[(view_ref, prop_ref)] = property_.source

                        # implicit end node type via container property, without actual knowledge of end node type
                        elif (
                            (
                                container := self.select_container(
                                    property_.container, property_.container_property_identifier
                                )
                            )
                            and (property_.container_property_identifier in container.properties)
                            and (
                                isinstance(
                                    container.properties[property_.container_property_identifier].type,
                                    DirectNodeRelation,
                                )
                            )
                        ):
                            connection_end_node_types[(view_ref, prop_ref)] = None

                    # reverse direct relation
                    elif isinstance(property_, ReverseDirectRelationProperty) and property_.source:
                        connection_end_node_types[(view_ref, prop_ref)] = property_.source

                    # edge property
                    elif isinstance(property_, EdgeProperty) and property_.source:
                        connection_end_node_types[(view_ref, prop_ref)] = property_.source

        return connection_end_node_types

    @staticmethod
    def is_explicit_connection(property_: ViewRequestProperty) -> bool:
        """Check if a property is an explicit connection property, meaning end node type is explicitly defined.

        Args:
            property_: The property to check.

        Returns:
            True if the property is a connection property, False otherwise.
        """
        return True if property_.source else False

    @cached_property
    def views_by_container(self) -> dict[ContainerReference, set[ViewReference]]:
        """Get a mapping from containers to the views that use them.

        Includes views from both the merged schema and all CDF views to capture
        container-view relationships across the entire CDF environment.
        Uses expanded views to include inherited properties.
        """

        # Include all unique views from merged and CDF
        views_by_container: dict[ContainerReference, set[ViewReference]] = defaultdict(set)

        # Include all unique views from merged and CDF
        all_view_refs = set(self.merged.views.keys()) | set(self.cdf.views.keys())

        for view_ref in all_view_refs:
            # Use expanded view to include inherited properties
            view = self.expand_view_properties(view_ref)
            if not view:
                continue

            for container in view.used_containers:
                views_by_container[container].add(view_ref)

        return dict(views_by_container)

    @cached_property
    def containers_by_view(self) -> dict[ViewReference, set[ContainerReference]]:
        """Get a mapping from views to the containers they use.

        Includes views from both the merged schema and all CDF views.
        Uses expanded views to include inherited properties.
        """
        containers_by_view: dict[ViewReference, set[ContainerReference]] = {}

        # Include all unique views from merged and CDF
        all_view_refs = set(self.merged.views.keys()) | set(self.cdf.views.keys())

        for view_ref in all_view_refs:
            # Use expanded view to include inherited properties
            view = self.expand_view_properties(view_ref)
            if view is not None:
                containers_by_view[view_ref] = view.used_containers

        return containers_by_view

    def find_views_mapping_to_containers(self, containers: list[ContainerReference]) -> set[ViewReference]:
        """Find views that map to all specified containers.

        That is, the intersection of views that use each of the specified containers.

        Args:
            containers: List of containers to check

        Returns:
            Set of views that contain all the specified containers

        Example:
            Given views V1, V2, V3 and containers C1, C2:
            - V1 uses containers {C1, C2}
            - V2 uses containers {C1}
            - V3 uses containers {C2}

            find_views_mapping_to_containers([C1, C2]) returns {V1}
            find_views_mapping_to_containers([C1]) returns {V1, V2}
            find_views_mapping_to_containers([C2]) returns {V1, V3}
        """
        if not containers:
            return set()

        view_sets = [self.views_by_container.get(c, set()) for c in containers]
        return set.intersection(*view_sets)

    @cached_property
    def implements_graph(self) -> nx.DiGraph:
        """Build a weighted directed graph of view implements.

        Nodes are ViewReferences, edges represent implements.
        An edge A → B means view A implements view B. Order of views in implements is used to set weight of an edge.

        Includes views from both merged schema and CDF
        """
        graph: nx.DiGraph = nx.DiGraph()

        for view_ref in self.cdf.views:
            graph.add_node(view_ref)
        for view_ref in self.merged.views:
            graph.add_node(view_ref)

        # Add edges for implements
        for view_ref in graph.nodes():
            view = self.select_view(view_ref)
            if not view or not view.implements:
                continue

            # Adding weight to preserve order of implements
            for i, implement in enumerate(view.implements):
                graph.add_edge(view_ref, implement, weight=i + 1)

        return graph

    @cached_property
    def implements_cycles(self) -> list[list[ViewReference]]:
        """Find all cycles in the implements graph.
        Returns:
            List of lists, where each list contains the ordered Views involved in forming the implements cycle.
        """

        return self.graph_cycles(self.implements_graph)

    @cached_property
    def requires_constraint_graph(self) -> nx.DiGraph:
        """Build a directed graph of container requires constraints.

        Nodes are ContainerReferences, edges represent requires constraints.
        An edge A → B means container A requires container B.

        Includes containers from both merged schema and CDF
        """
        graph: nx.DiGraph = nx.DiGraph()

        for container_ref in self.cdf.containers:
            graph.add_node(container_ref)
        for container_ref in self.merged.containers:
            graph.add_node(container_ref)

        # Add edges for requires constraints from all known containers
        for container_ref in list(graph.nodes()):
            container = self.select_container(container_ref)
            if not container or not container.constraints:
                continue
            for constraint_id, constraint in container.constraints.items():
                if not isinstance(constraint, RequiresConstraintDefinition):
                    continue
                is_auto = constraint_id.endswith("__auto")
                graph.add_edge(container_ref, constraint.require, is_auto=is_auto)

        return graph

    @cached_property
    def modifiable_containers(self) -> set[ContainerReference]:
        """Containers whose requires constraints can be modified in this session.

        A container is modifiable if:
        - It's NOT in a CDF built-in space (CDM, IDM, etc.)
        - It's a user container brought in through the loaded data model scope or view implements chain
        """
        return {container_ref for container_ref in self.merged.containers if container_ref.space not in COGNITE_SPACES}

    @cached_property
    def immutable_requires_constraint_graph(self) -> nx.DiGraph:
        """Subgraph of requires constraints from non-modifiable (CDM) containers.

        Used to check reachability via existing immutable constraints.
        """
        return nx.subgraph_view(
            self.requires_constraint_graph,
            filter_edge=lambda src, _: src not in self.modifiable_containers,
        )

    @cached_property
    def _fixed_constraint_graph(self) -> nx.DiGraph:
        """Graph of all fixed constraints (immutable + user-intentional).

        Both are "fixed" from the optimizer's perspective - existing paths that can't be changed.
        """
        G = nx.DiGraph()
        G.add_edges_from(self.immutable_requires_constraint_graph.edges())
        G.add_edges_from(self._user_intentional_constraints)
        return G

    @cached_property
    def _fixed_descendants(self) -> defaultdict[ContainerReference, set[ContainerReference]]:
        """Pre-compute descendants via fixed constraints. Missing keys return empty set."""
        result: defaultdict[ContainerReference, set[ContainerReference]] = defaultdict(set)
        for container in self._fixed_constraint_graph.nodes():
            result[container] = nx.descendants(self._fixed_constraint_graph, container)
        return result

    @cached_property
    def _existing_requires_edges(self) -> set[tuple[ContainerReference, ContainerReference]]:
        """Cached set of existing requires constraint edges."""
        return set(self.requires_constraint_graph.edges())

    @cached_property
    def _user_intentional_constraints(self) -> set[tuple[ContainerReference, ContainerReference]]:
        """Constraints that appear to be user-intentional and should not be auto-removed

        A constraint is user-intentional if:
        1. The constraint identifier does NOT have '__auto' postfix
        2. Neither src nor dst is part of a cycle (cyclic constraints are errors)

        These constraints are preserved even if they're not in the optimal structure, because
        they may be used for data integrity purposes.
        We DON'T consider manual-created constraints as user-intended if they form part of a cycle,
        because that indicates a problem with the data model where we likely can provide a better solution.
        """
        containers_in_cycles = {container for cycle in self.requires_constraint_cycles for container in cycle}

        return {
            (src, dst)
            for src, dst, data in self.requires_constraint_graph.edges(data=True)
            if not data.get("is_auto", False) and src not in containers_in_cycles and dst not in containers_in_cycles
        }

    @staticmethod
    def forms_directed_path(nodes: set[_NodeT], graph: nx.DiGraph) -> bool:
        """Check if nodes form an uninterrupted directed path in the graph.

        Returns True if there exists a node that can reach all other nodes via
        directed edges in the graph.

        Args:
            nodes: Set of nodes to check
            graph: Directed graph containing the nodes

        Returns:
            True if nodes form a directed path (one node reaches all others)

        Example:
            Given nodes N1, N2, N3 with edges:
            - N1 -> N2
            - N2 -> N3

            forms_directed_path({N1, N2, N3}) returns True (N1 reaches all others)
            forms_directed_path({N2, N3}) returns True (N2 reaches N3)
            forms_directed_path({N1, N3}) returns False (N1 can't reach N3 without N2)
        """
        if len(nodes) <= 1:
            return True

        # Filter to nodes that are actually in the graph
        nodes_in_graph = {n for n in nodes if n in graph}
        if len(nodes_in_graph) < len(nodes):
            # Some nodes aren't in the graph, so we can't form a complete path
            return False

        for candidate in nodes_in_graph:
            others = nodes_in_graph - {candidate}
            if others.issubset(nx.descendants(graph, candidate)):
                return True

        return False

    @cached_property
    def requires_constraint_cycles(self) -> list[list[ContainerReference]]:
        """Find all cycles in the requires constraint graph.
        Returns:
            List of lists, where each list contains the ordered containers involved in forming the requires cycle.
        """
        return self.graph_cycles(self.requires_constraint_graph)

    @staticmethod
    def graph_cycles(graph: nx.DiGraph) -> list[list[T_Reference]]:
        """Returns cycles in the graph otherwise empty list"""
        return [candidate for candidate in nx.simple_cycles(graph) if len(candidate) > 1]

    @cached_property
    def optimized_requires_constraint_graph(self) -> nx.DiGraph:
        """Target state of requires constraints after optimizing (MST + fixed constraints)."""
        graph = nx.DiGraph()
        graph.add_edges_from(self.oriented_mst_edges)
        graph.add_edges_from(self._fixed_constraint_graph.edges())
        return graph

    @cached_property
    def _requires_candidate_graph(self) -> nx.Graph:
        """Build weighted candidate graph for requires constraints.

        Contains all container pairs that appear together in any view, with:
        - weight: minimum directional weight (for MST computation)
        - preferred_direction: direction with lower weight (for tie-breaking)

        This graph is used to compute per-view MSTs.
        """
        G = nx.Graph()

        for view_ref in self.merged.views:
            containers = self.containers_by_view.get(view_ref, set())
            if len(containers) < 2:
                continue  # Need at least 2 containers to form a requires constraint

            # Sort for deterministic preferred_direction when weights are equal
            for src, dst in combinations(sorted(containers, key=str), 2):
                if G.has_edge(src, dst):
                    continue  # Already added from another view

                w1 = self._compute_requires_edge_weight(src, dst)
                w2 = self._compute_requires_edge_weight(dst, src)
                direction = (src, dst) if w1 <= w2 else (dst, src)
                weight = min(w1, w2)

                G.add_edge(src, dst, weight=weight, preferred_direction=direction)

        return G

    @cached_property
    def _mst_by_view(self) -> dict[ViewReference, nx.Graph]:
        """Compute per-view MST graphs.

        Each view gets its own MST over just its containers. This ensures:
        - No routing through containers not in the view
        - Each view gets exactly the edges it needs
        - Voting handles orientation conflicts between views

        Skips inherently unsolvable views (no immutable anchor + all modifiables are roots).
        """
        if not self._requires_candidate_graph:
            return {}

        result: dict[ViewReference, nx.Graph] = {}

        for view_ref in self.merged.views:
            if view_ref in self._views_with_root_conflicts:
                continue

            containers = self.containers_by_view.get(view_ref, set())
            if len(containers) < 2:  # Need at least 2 containers to have a constraint
                continue
            if not containers.intersection(self.modifiable_containers):
                continue

            subgraph = self._requires_candidate_graph.subgraph(containers)
            if not nx.is_connected(subgraph):
                continue

            result[view_ref] = nx.minimum_spanning_tree(subgraph, weight="weight")

        return result

    @cached_property
    def _root_by_view(self) -> dict[ViewReference, ContainerReference]:
        """Map each view (with 2+ containers) to its most view-specific (root) container.

        Only includes views with 2+ containers since single-container views
        are trivially satisfied and don't need root direction.

        Selection criteria (in priority order):
        1. Fewest views: Containers appearing in fewer views are more "view-specific"
        2. Has existing constraint: Prefer containers with existing outgoing constraints
        3. Alphabetical: Deterministic tie-breaker
        """
        result: dict[ViewReference, ContainerReference] = {}

        for view, containers in self.containers_by_view.items():
            if len(containers) < 2:
                continue

            modifiable = containers.intersection(self.modifiable_containers)
            if not modifiable:
                continue

            # Selection priority: fewest views, has existing constraint, alphabetical
            result[view] = min(
                modifiable,
                key=lambda c: (
                    len(self.views_by_container.get(c, set())),
                    0 if any((c, other) in self._existing_requires_edges for other in containers) else 1,
                    str(c),
                ),
            )

        return result

    @cached_property
    def _views_with_root_conflicts(self) -> set[ViewReference]:
        """Views where all modifiable containers are forced roots for other views.

        A view has root conflicts if:
        1. It has no immutable containers (no shared CDM anchor)
        2. All its modifiable containers are already forced roots for other views

        Such views would require edges between forced roots, causing conflicts.
        """
        forced_roots = set(self._root_by_view.values())
        unsolvable: set[ViewReference] = set()

        for view, containers in self.containers_by_view.items():
            modifiable = containers & self.modifiable_containers
            immutable = containers - self.modifiable_containers

            # Need at least 2 modifiable containers to have a conflict
            if len(modifiable) < 2:
                continue

            # No immutable anchor AND all modifiables are roots elsewhere
            if not immutable and modifiable <= forced_roots:
                unsolvable.add(view)

        return unsolvable

    @cached_property
    def oriented_mst_edges(self) -> set[tuple[ContainerReference, ContainerReference]]:
        """Orient per-view MST edges by voting across views.

        Each view votes for edge orientations based on BFS from its root container.
        Views with only 1 modifiable container use 'inf' vote weight to force
        that container as root.

        Tie-breaker: preferred_direction from weight function.

        Returns set of directed (src, dst) tuples.
        """
        edge_votes: dict[tuple[ContainerReference, ContainerReference], float] = defaultdict(float)
        all_edges: set[tuple[ContainerReference, ContainerReference]] = set()

        # Sort for deterministic iteration (dict order can vary with hash randomization)
        for view in sorted(self._mst_by_view.keys(), key=str):
            mst = self._mst_by_view[view]
            root = self._root_by_view[view]  # Always exists for views in _mst_by_view
            containers = self.containers_by_view.get(view, set())
            modifiable_count = len(containers & self.modifiable_containers)
            # Views with only 1 modifiable container have no choice - that container MUST be root
            vote_weight = float("inf") if modifiable_count == 1 else 1.0

            # BFS from root orients edges away from root (parent → child)
            for parent, child in nx.bfs_edges(mst, root):  # type: ignore[func-returns-value]
                if parent in self.modifiable_containers:
                    edge_votes[(parent, child)] += vote_weight

            # Normalize edges to canonical form so votes for same undirected edge are counted together
            for c1, c2 in mst.edges():
                all_edges.add((c1, c2) if str(c1) < str(c2) else (c2, c1))

        # Pick direction: most votes wins, preferred_direction breaks ties
        oriented: set[tuple[ContainerReference, ContainerReference]] = set()

        # Sort for deterministic iteration (hash randomization affects set order)
        for c1, c2 in sorted(all_edges, key=lambda e: (str(e[0]), str(e[1]))):
            c1_votes = edge_votes.get((c1, c2), 0)
            c2_votes = edge_votes.get((c2, c1), 0)

            if c1_votes > c2_votes:
                oriented.add((c1, c2))
            elif c2_votes > c1_votes:
                oriented.add((c2, c1))
            else:
                # Tie-breaker: use preferred_direction from weight function
                oriented.add(self._requires_candidate_graph[c1][c2].get("preferred_direction", (c1, c2)))

        return oriented

    @cached_property
    def _transitively_reduced_edges(self) -> set[tuple[ContainerReference, ContainerReference]]:
        """Reduce MST edges to minimal necessary set (remove edges with alternative paths via immutable)."""
        if not self.oriented_mst_edges:
            return set()

        # Optimal graph = MST + immutable + user-intentional (these provide existing paths)
        optimal = nx.DiGraph()
        optimal.add_edges_from(self.immutable_requires_constraint_graph.edges())
        optimal.add_edges_from(self._user_intentional_constraints)
        optimal.add_edges_from(self.oriented_mst_edges)

        reduced = nx.transitive_reduction(optimal)

        # Return MST edges that survive reduction
        return {e for e in reduced.edges() if e in self.oriented_mst_edges}

    def get_requires_changes_for_view(self, view: ViewReference) -> RequiresChangesForView:
        """Get requires constraint changes needed to optimize a view.

        Returns a RequiresChangesForView with:
        - to_add: New constraints needed where source is mapped in this view
        - to_remove: Existing constraints that are redundant or wrongly oriented
        - status: The optimization status for this view
        """
        containers_in_view = self.containers_by_view.get(view, set())
        modifiable_containers_in_view = containers_in_view.intersection(self.modifiable_containers)
        if not modifiable_containers_in_view:
            # No modifiable containers - check if view is already optimized via existing constraints
            if self.forms_directed_path(containers_in_view, self.requires_constraint_graph):
                return RequiresChangesForView(set(), set(), RequiresChangeStatus.OPTIMAL)
            return RequiresChangesForView(set(), set(), RequiresChangeStatus.NO_MODIFIABLE_CONTAINERS)

        # Early exit for inherently unsolvable views (no CDM anchor + all modifiables are roots)
        if view in self._views_with_root_conflicts:
            return RequiresChangesForView(
                set[tuple[ContainerReference, ContainerReference]](), set(), RequiresChangeStatus.UNSOLVABLE
            )

        # Filter edges to those where source is in this view's modifiable containers
        existing_from_view = {
            edge for edge in self._existing_requires_edges if edge[0] in modifiable_containers_in_view
        }
        optimal_for_view = {
            edge for edge in self._transitively_reduced_edges if edge[0] in modifiable_containers_in_view
        }

        to_add = optimal_for_view - existing_from_view

        # To remove: existing edges with wrong direction or not in MST (and not needed externally)
        # But NEVER remove user-intentional constraints (manually defined, no __auto postfix)
        to_remove: set[tuple[ContainerReference, ContainerReference]] = set()
        for src, dst in existing_from_view:
            # Skip user-intentional constraints - they were set by the user on purpose
            if (src, dst) in self._user_intentional_constraints:
                continue
            if (dst, src) in self.oriented_mst_edges:
                to_remove.add((src, dst))  # Always remove if opposite direction from optimal solution
            elif (src, dst) not in self.oriented_mst_edges and not (
                self.find_views_mapping_to_containers([src, dst]) - set(self.merged.views)
            ):
                to_remove.add((src, dst))  # Remove if not in optimal solution and not needed by external views

        # Check solvability in optimized state
        if not self.forms_directed_path(
            self.containers_by_view.get(view, set()), self.optimized_requires_constraint_graph
        ):
            return RequiresChangesForView(set(), set(), RequiresChangeStatus.UNSOLVABLE)

        if not to_add and not to_remove:
            return RequiresChangesForView(set(), set(), RequiresChangeStatus.OPTIMAL)

        return RequiresChangesForView(to_add, to_remove, RequiresChangeStatus.CHANGES_AVAILABLE)

    # ========================================================================
    # REQUIRES CONSTRAINT MST WEIGHT CONSTANTS
    # ========================================================================
    # Weight = TIER + sub_weight.
    #
    # WHY THIS ENCODING:
    # MST algorithm only support scalar weights, but we need a strict priority
    # hierarchy where USER→USER ALWAYS beats USER→EXTERNAL. By using a large
    # gap (1000) between tiers, sub-weights (max ~50) can never cause a lower
    # tier to beat a higher tier.
    #
    # Tiers (explicit priority order):
    #   - Tier 1 (USER→USER):     Both containers modifiable - always preferred
    #   - Tier 2 (USER→EXTERNAL): Target is CDF/CDM - only when needed
    #   - Tier ∞ (FORBIDDEN):     Invalid edge, forms cycle or source is not modifiable
    #
    # Sub-weights refine ordering WITHIN a tier (shared views, direction, etc).
    #   - These have been empirically tuned through trial and error.
    # ========================================================================

    # Tier base weights (gap of 1000 ensures tier always wins)
    _TIER_USER_TO_USER = 1000
    _TIER_USER_TO_EXTERNAL = 2000
    _TIER_FORBIDDEN = math.inf

    # Sub-weight adjustments (applied within tier, max ~100)
    _BONUS_SHARED_VIEWS_PER = 5  # Per shared view (max 5 views → 25)
    _BONUS_SHARED_VIEWS_MAX = 25
    _BONUS_COVERAGE_PER = 5  # Per descendant via immutable edges (max 3 → 15)
    _BONUS_COVERAGE_MAX = 15
    _PENALTY_VIEW_COUNT = 10  # When src is in more views than dst

    # Tie-breaker for deterministic ordering
    _TIE_BREAKER_DIVISOR = 1e9

    def _compute_requires_edge_weight(self, src: ContainerReference, dst: ContainerReference) -> float:
        """Compute the weight/cost of adding edge src → dst.

        Returns TIER + sub_weight where tier dominates (gap of 1000).
        Sub-weights refine ordering within a tier based on shared views, direction, coverage.
        """
        # Opposite direction of fixed constraints is forbidden (would conflict with existing path)
        if src in self._fixed_descendants[dst]:
            return self._TIER_FORBIDDEN

        if src not in self.modifiable_containers:
            return self._TIER_FORBIDDEN

        src_views = self.views_by_container.get(src, set())
        dst_views = self.views_by_container.get(dst, set())

        # Sub-weight adjustments
        shared_bonus = min(len(src_views & dst_views) * self._BONUS_SHARED_VIEWS_PER, self._BONUS_SHARED_VIEWS_MAX)
        coverage_bonus = min(len(self._fixed_descendants[dst]) * self._BONUS_COVERAGE_PER, self._BONUS_COVERAGE_MAX)
        view_penalty = self._PENALTY_VIEW_COUNT if len(src_views) > len(dst_views) else 0

        # Deterministic tie-breaker (very small, only matters when all else is equal)
        edge_str = f"{src.space}:{src.external_id}->{dst.space}:{dst.external_id}"
        tie_breaker = sum(ord(c) for c in edge_str) / self._TIE_BREAKER_DIVISOR

        if dst in self.modifiable_containers:
            return self._TIER_USER_TO_USER - shared_bonus - coverage_bonus + view_penalty + tie_breaker

        return self._TIER_USER_TO_EXTERNAL - shared_bonus - coverage_bonus + tie_breaker



import inspect
from abc import ABC, abstractmethod
from functools import lru_cache
from types import UnionType
from typing import Generic, TypeVar, Union, get_args, get_origin

from cognite.neat._v0.core._constants import DEFAULT_NAMESPACE
from cognite.neat._v0.core._data_model._shared import (
    DataModel,
    ImportedDataModel,
    VerifiedDataModel,
)
from cognite.neat._v0.core._data_model.models import (
    UnverifiedConceptualDataModel,
    UnverifiedPhysicalDataModel,
)
from cognite.neat._v0.core._store._provenance import Agent as ProvenanceAgent

T_DataModelIn = TypeVar("T_DataModelIn", bound=DataModel)
T_DataModelOut = TypeVar("T_DataModelOut", bound=DataModel)
T_VerifiedIn = TypeVar("T_VerifiedIn", bound=VerifiedDataModel)
T_VerifiedOut = TypeVar("T_VerifiedOut", bound=VerifiedDataModel)


class DataModelTransformer(ABC, Generic[T_DataModelIn, T_DataModelOut]):
    """This is the base class for all data model transformers."""

    @abstractmethod
    def transform(self, data_model: T_DataModelIn) -> T_DataModelOut:
        """Transform the input data model into the output data model."""
        raise NotImplementedError()

    @property
    def agent(self) -> ProvenanceAgent:
        """Provenance agent for the importer."""
        return ProvenanceAgent(id_=DEFAULT_NAMESPACE[f"agent/{type(self).__name__}"])

    @property
    def description(self) -> str:
        """Get the description of the transformer."""
        return "MISSING DESCRIPTION"

    def is_valid_input(self, data_model: T_DataModelIn) -> bool:
        """Check if the input data model is valid."""
        types = self.transform_type_hint()
        for type_ in types:
            if get_origin(type_) is ImportedDataModel:
                inner = get_args(type_)[0]
                if isinstance(data_model, ImportedDataModel) and isinstance(data_model.unverified_data_model, inner):
                    return True
            elif isinstance(data_model, type_):
                return True
        return False

    @classmethod
    @lru_cache(maxsize=1)
    def transform_type_hint(cls) -> tuple[type, ...]:
        # This is an expensive operation, so we cache the result
        signature = inspect.signature(cls.transform)
        annotation = signature.parameters["data_model"].annotation
        if isinstance(annotation, TypeVar):
            if annotation.__bound__ is None:
                raise TypeError(f"TypeVar {annotation} must be bound to a type.")
            annotation = annotation.__bound__
        # The annotation can be a type or a generic
        if get_origin(annotation) in [UnionType, Union]:
            return get_args(annotation)

        if get_origin(annotation) is ImportedDataModel and isinstance(get_args(annotation)[0], TypeVar):
            # Hardcoded for now, as we only have two types of imported data models
            return (
                ImportedDataModel[UnverifiedPhysicalDataModel],
                ImportedDataModel[UnverifiedConceptualDataModel],
            )

        return (annotation,)


class VerifiedDataModelTransformer(DataModelTransformer[T_VerifiedIn, T_VerifiedOut], ABC): ...



import re
from abc import ABC
from typing import Any, Literal

from pydantic import Field, field_validator
from pydantic_core.core_schema import ValidationInfo

from cognite.neat._utils.text import humanize_collection
from cognite.neat._utils.useful_types import BaseModelObject

from ._base import APIResource, Resource, WriteableResource
from ._constants import (
    CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    DM_EXTERNAL_ID_PATTERN,
    FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS,
    FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER,
    SPACE_FORMAT_PATTERN,
)
from ._constraints import Constraint
from ._data_types import DataType
from ._indexes import Index
from ._references import ContainerReference

KEY_PATTERN = re.compile(CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN)


class ContainerPropertyDefinition(BaseModelObject):
    immutable: bool | None = Field(
        default=None,
        description="Should updates to this property be rejected after the initial population?",
    )
    nullable: bool | None = Field(
        default=None,
        description="Does this property need to be set to a value, or not?",
    )
    auto_increment: bool | None = Field(
        default=None,
        description="Increment the property based on its highest current value (max value).",
    )
    default_value: str | int | float | bool | dict[str, Any] | None = Field(
        default=None,
        description="Default value to use when you do not specify a value for the property.",
    )
    description: str | None = Field(
        default=None,
        description="Description of the content and suggested use for this property.",
        max_length=1024,
    )
    name: str | None = Field(
        default=None,
        description="Readable property name.",
        max_length=255,
    )
    type: DataType = Field(description="The type of data you can store in this property.")


class Container(Resource, APIResource[ContainerReference], ABC):
    space: str = Field(
        description="The workspace for the container, a unique identifier for the space.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the container.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    name: str | None = Field(
        default=None,
        description="name for the container.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the container.",
        max_length=1024,
    )
    used_for: Literal["node", "edge", "all"] | None = Field(
        default=None,
        description="Should this operation apply to nodes, edges or both.",
    )
    properties: dict[str, ContainerPropertyDefinition] = Field(
        description="Set of properties to apply to the container.",
    )
    constraints: dict[str, Constraint] | None = Field(
        default=None,
        description="Set of constraints to apply to the container.",
        max_length=10,
    )
    indexes: dict[str, Index] | None = Field(
        default=None,
        description="Set of indexes to apply to the container.",
        max_length=10,
    )

    @field_validator("indexes", "constraints", mode="after")
    def validate_key_length(cls, val: dict[str, Any] | None, info: ValidationInfo) -> dict[str, Any] | None:
        """Validate keys"""
        if not isinstance(val, dict):
            return val
        invalid_keys = {key for key in val.keys() if not (1 <= len(key) <= 43)}
        if invalid_keys:
            raise ValueError(
                f"{info.field_name} keys must be between 1 and 43 characters long. Invalid keys: "
                f"{humanize_collection(invalid_keys)}"
            )
        return val

    @field_validator("properties", mode="after")
    def validate_property_keys(cls, val: dict[str, Any]) -> dict[str, Any]:
        """Validate property keys"""
        if invalid_keys := {key for key in val if not KEY_PATTERN.match(key)}:
            raise ValueError(
                f"Property keys must match pattern '{CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN}'. "
                f"Invalid keys: {humanize_collection(invalid_keys)}"
            )

        if forbidden_keys := set(val.keys()).intersection(FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER):
            raise ValueError(
                f"Property keys cannot be any of the following reserved values: {humanize_collection(forbidden_keys)}"
            )
        return val

    @field_validator("external_id")
    def check_forbidden_external_id_value(cls, val: str) -> str:
        """Check the external_id not present in forbidden set"""
        if val in FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS:
            raise ValueError(
                f"{val!r} is a reserved container External ID. Reserved External IDs are: "
                f"{humanize_collection(FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS)}"
            )
        return val

    def as_reference(self) -> ContainerReference:
        return ContainerReference(
            space=self.space,
            external_id=self.external_id,
        )


class ContainerRequest(Container): ...


class ContainerResponse(Container, WriteableResource[ContainerRequest]):
    created_time: int = Field(
        description="When the container was created. The number of milliseconds since 00:00:00 "
        "Thursday, 1 January 1970, "
        "Coordinated Universal Time (UTC), minus leap seconds."
    )
    last_updated_time: int = Field(
        description="When the container was last updated. The number of milliseconds since 00:00:00 Thursday, "
        "1 January 1970, Coordinated Universal Time (UTC), minus leap seconds."
    )
    is_global: bool = Field(description="Whether the container is a global container.")

    def as_request(self) -> "ContainerRequest":
        return ContainerRequest.model_validate(self.model_dump(by_alias=True))



from pathlib import Path
from typing import Protocol, TypeVar

from cognite.neat._v0.core._issues import NeatIssue
from cognite.neat._v0.core._issues.errors import (
    FileNotAFileError,
    FileNotFoundNeatError,
    FileTypeUnexpectedError,
)

T = TypeVar("T")
DataT = TypeVar("DataT", covariant=True)


class FileReader(Protocol[DataT]):
    """Protocol for file readers that parse files into structured data."""

    def read_file(self, filepath: Path) -> tuple[DataT, list[NeatIssue]]:
        """Read a file and return the data with any issues encountered.

        Args:
            filepath: Path to the file to read

        Returns:
            Tuple of (parsed_data, issues_list)
        """
        ...


class BaseFileReader:
    """Base implementation with common file validation logic."""

    @staticmethod
    def validate_file(filepath: Path, allowed_extensions: frozenset[str]) -> list[NeatIssue]:
        """Validate that a file exists and has the correct extension.

        Args:
            filepath: Path to the file to validate
            allowed_extensions: Set of allowed file extensions

        Returns:
            List of issues found during validation, empty if valid
        """
        # Check if file exists
        if not filepath.exists():
            return [FileNotFoundNeatError(filepath)]

        # Check if it's a file
        if not filepath.is_file():
            return [FileNotAFileError(filepath)]

        # Check file extension
        if filepath.suffix not in allowed_extensions:
            return [FileTypeUnexpectedError(filepath, allowed_extensions)]

        return []



import itertools
import urllib.parse
import warnings
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import cast

from cognite.client.data_classes.data_modeling import DataModelId, DataModelIdentifier
from rdflib import RDF, RDFS, Namespace, URIRef
from rdflib import Literal as RdfLiteral
from rdflib.query import ResultRow

from cognite.neat._v0.core._config import GLOBAL_CONFIG
from cognite.neat._v0.core._constants import NEAT
from cognite.neat._v0.core._data_model._shared import ImportedDataModel
from cognite.neat._v0.core._data_model.models import UnverifiedConceptualDataModel
from cognite.neat._v0.core._data_model.models._import_contexts import GraphContext
from cognite.neat._v0.core._data_model.models.conceptual import (
    UnverifiedConcept,
    UnverifiedConceptualMetadata,
    UnverifiedConceptualProperty,
)
from cognite.neat._v0.core._data_model.models.entities import UnknownEntity
from cognite.neat._v0.core._issues.errors import NeatValueError
from cognite.neat._v0.core._issues.warnings import NeatValueWarning
from cognite.neat._v0.core._store import NeatInstanceStore
from cognite.neat._v0.core._utils.collection_ import iterate_progress_bar
from cognite.neat._v0.core._utils.rdf_ import split_uri

from ._base import BaseImporter


# Internal helper class
@dataclass
class _ReadProperties:
    type_uri: URIRef
    property_uri: URIRef
    value_type: URIRef
    parent_uri: URIRef | None
    max_occurrence: int
    instance_count: int


class GraphImporter(BaseImporter[UnverifiedConceptualDataModel]):
    """Infers a data model from the triples in a Graph Store.
    Args:
        store: The graph store to read from.
        data_model_id: The data model id to be used for the imported rules.
    """

    _ORDERED_CONCEPTS_QUERY = """SELECT DISTINCT ?concept (count(?s) as ?instances )
                           WHERE { ?s a ?concept }
                           group by ?concept order by DESC(?instances)"""

    _TYPE_PARENT_QUERY = f"""SELECT ?parent ?type
                            WHERE {{ ?s a ?type .
                            ?type <{RDFS.subClassOf}> ?parent }}"""

    _PROPERTIES_QUERY = """SELECT DISTINCT ?property ?valueType
                         WHERE {{
                            ?s a <{type}> .
                            ?s ?property ?object .
                            OPTIONAL {{ ?object a ?objectType }}
                            BIND(
                               IF(
                                    isLiteral(?object), datatype(?object),
                                    IF(BOUND(?objectType), ?objectType, <{unknown_type}>)
                                ) AS ?valueType
                            )
                        }}"""

    _MAX_OCCURRENCE_QUERY = """SELECT (MAX(?count) AS ?maxCount)
                            WHERE {{
                              {{
                                SELECT ?subject (COUNT(?object) AS ?count)
                                WHERE {{
                                  ?subject a <{type}> .
                                  ?subject <{property}> ?object .
                                }}
                                GROUP BY ?subject
                              }}
                            }}"""

    def __init__(
        self,
        store: NeatInstanceStore,
        data_model_id: DataModelIdentifier = ("neat_space", "NeatInferredDataModel", "v1"),
    ) -> None:
        self.store = store
        self.data_model_id = DataModelId.load(data_model_id)
        if self.data_model_id.version is None:
            raise NeatValueError("Version is required when setting a Data Model ID")

    def to_data_model(self) -> ImportedDataModel[UnverifiedConceptualDataModel]:
        metadata = self._create_default_metadata()
        if not self.store.queries.select.has_data():
            warnings.warn(NeatValueWarning("Cannot infer data model. No data found in the graph."), stacklevel=2)
            return ImportedDataModel(UnverifiedConceptualDataModel(metadata, [], [], {}), None)

        parent_by_child = self._read_parent_by_child_from_graph()
        count_by_type = self._read_types_with_counts_from_graph()
        if not count_by_type:
            warnings.warn(
                NeatValueWarning("Cannot infer data model. No RDF.type triples found in the graph."), stacklevel=2
            )
            return ImportedDataModel(UnverifiedConceptualDataModel(metadata, [], [], {}), None)

        read_properties = self._read_concept_properties_from_graph(count_by_type, parent_by_child)

        prefixes: dict[str, Namespace] = {}
        concepts, properties = self._create_concepts_properties(read_properties, prefixes)

        return ImportedDataModel(
            UnverifiedConceptualDataModel(
                metadata=metadata,
                concepts=concepts,
                properties=properties,
                prefixes=prefixes,
            ),
            context=GraphContext({"inferred_from": count_by_type}),
        )

    def _read_parent_by_child_from_graph(self) -> dict[URIRef, URIRef]:
        parent_by_child: dict[URIRef, URIRef] = {}
        for result_row in self.store.dataset.query(self._TYPE_PARENT_QUERY):
            parent_uri, child_uri = cast(tuple[URIRef, URIRef], result_row)
            parent_by_child[child_uri] = parent_uri
        return parent_by_child

    def _read_types_with_counts_from_graph(self) -> dict[URIRef, int]:
        count_by_type: dict[URIRef, int] = {}
        # Reads all types and their instance counts from the graph
        for result_row in self.store.dataset.query(self._ORDERED_CONCEPTS_QUERY):
            type_uri, instance_count_literal = cast(tuple[URIRef, RdfLiteral], result_row)
            count_by_type[type_uri] = instance_count_literal.toPython()
        return count_by_type

    def _read_concept_properties_from_graph(
        self, count_by_type: dict[URIRef, int], parent_by_child: dict[URIRef, URIRef]
    ) -> list[_ReadProperties]:
        read_properties: list[_ReadProperties] = []

        total_instance_count = sum(count_by_type.values())
        iterable = count_by_type.items()
        if GLOBAL_CONFIG.use_iterate_bar_threshold and total_instance_count > GLOBAL_CONFIG.use_iterate_bar_threshold:
            iterable = iterate_progress_bar(iterable, len(count_by_type), "Inferring types...")  # type: ignore[assignment]

        for type_uri, instance_count in iterable:
            property_query = self._PROPERTIES_QUERY.format(type=type_uri, unknown_type=NEAT.UnknownType)
            for result_row in self.store.dataset.query(property_query):
                property_uri, value_type_uri = cast(tuple[URIRef, URIRef], result_row)
                if property_uri == RDF.type:
                    continue
                occurrence_query = self._MAX_OCCURRENCE_QUERY.format(type=type_uri, property=property_uri)
                max_occurrence = 1  # default value
                # We know that the _MAX_OCCURRENCE_QUERY will return a ResultRow
                occurrence_results = list(cast(ResultRow, self.store.dataset.query(occurrence_query)))
                if occurrence_results and occurrence_results[0] and occurrence_results[0][0]:
                    max_occurrence_literal = cast(RdfLiteral, occurrence_results[0][0])
                    max_occurrence = int(max_occurrence_literal.toPython())
                read_properties.append(
                    _ReadProperties(
                        type_uri=type_uri,
                        property_uri=property_uri,
                        parent_uri=parent_by_child.get(type_uri),
                        value_type=value_type_uri,
                        max_occurrence=max_occurrence,
                        instance_count=instance_count,
                    )
                )
        return read_properties

    def _create_concepts_properties(
        self, read_properties: list[_ReadProperties], prefixes: dict[str, Namespace]
    ) -> tuple[list[UnverifiedConcept], list[UnverifiedConceptualProperty]]:
        concepts: list[UnverifiedConcept] = []
        properties: list[UnverifiedConceptualProperty] = []

        # Help for IDE
        type_uri: URIRef
        parent_uri: URIRef
        for parent_uri, parent_concepts_properties_iterable in itertools.groupby(
            sorted(read_properties, key=lambda x: x.parent_uri or NEAT.EmptyType),
            key=lambda x: x.parent_uri or NEAT.EmptyType,
        ):
            parent_str: str | None = None
            if parent_uri != NEAT.EmptyType:
                parent_str, parent_cls = self._create_concept(parent_uri, set_instance_source=False, prefixes=prefixes)
                concepts.append(parent_cls)

            properties_by_concept_by_property = self._get_properties_by_concept_by_property(
                parent_concepts_properties_iterable
            )
            for type_uri, properties_by_property_uri in properties_by_concept_by_property.items():
                concept_str, concept = self._create_concept(
                    type_uri, set_instance_source=True, prefixes=prefixes, implements=parent_str
                )
                concepts.append(concept)
                for property_uri, read_properties in properties_by_property_uri.items():
                    namespace, property_suffix = split_uri(property_uri)
                    (self._add_uri_namespace_to_prefixes(namespace, prefixes),)
                    properties.append(
                        self._create_property(
                            read_properties, concept_str, property_uri, urllib.parse.unquote(property_suffix), prefixes
                        )
                    )
        return concepts, properties

    @staticmethod
    def _get_properties_by_concept_by_property(
        parent_concept_properties_iterable: Iterable[_ReadProperties],
    ) -> dict[URIRef, dict[URIRef, list[_ReadProperties]]]:
        properties_by_concept_by_property: dict[URIRef, dict[URIRef, list[_ReadProperties]]] = {}
        for concept_uri, concept_properties_iterable in itertools.groupby(
            sorted(parent_concept_properties_iterable, key=lambda x: x.type_uri), key=lambda x: x.type_uri
        ):
            properties_by_concept_by_property[concept_uri] = defaultdict(list)
            for read_prop in concept_properties_iterable:
                properties_by_concept_by_property[concept_uri][read_prop.property_uri].append(read_prop)
        return properties_by_concept_by_property

    @classmethod
    def _create_concept(
        cls, type_uri: URIRef, set_instance_source: bool, prefixes: dict[str, Namespace], implements: str | None = None
    ) -> tuple[str, UnverifiedConcept]:
        namespace, suffix = split_uri(type_uri)
        cls._add_uri_namespace_to_prefixes(namespace, prefixes)
        concept_str = urllib.parse.unquote(suffix)
        return concept_str, UnverifiedConcept(
            concept=concept_str, implements=implements, instance_source=type_uri if set_instance_source else None
        )

    def _create_property(
        self,
        read_properties: list[_ReadProperties],
        concept_str: str,
        property_uri: URIRef,
        property_id: str,
        prefixes: dict[str, Namespace],
    ) -> UnverifiedConceptualProperty:
        first = read_properties[0]
        value_type = self._get_value_type(read_properties, prefixes)
        return UnverifiedConceptualProperty(
            concept=concept_str,
            property_=property_id,
            max_count=first.max_occurrence,
            value_type=value_type,
            instance_source=[property_uri],
        )

    @classmethod
    def _get_value_type(
        cls, read_properties: list[_ReadProperties], prefixes: dict[str, Namespace]
    ) -> str | UnknownEntity:
        value_types = {prop.value_type for prop in read_properties}
        if len(value_types) == 1:
            uri_ref = value_types.pop()
            if uri_ref == NEAT.UnknownType:
                return UnknownEntity()
            namespace, suffix = split_uri(uri_ref)
            cls._add_uri_namespace_to_prefixes(namespace, prefixes)
            return suffix
        uri_refs: list[str] = []
        for uri_ref in value_types:
            if uri_ref == NEAT.UnknownType:
                return UnknownEntity()
            namespace, suffix = split_uri(uri_ref)
            cls._add_uri_namespace_to_prefixes(namespace, prefixes)
            uri_refs.append(suffix)
        # Sort the URIs to ensure deterministic output
        return ", ".join(sorted(uri_refs))

    def _create_default_metadata(self) -> UnverifiedConceptualMetadata:
        now = datetime.now(timezone.utc)
        name = self.data_model_id.external_id.replace("_", " ").title()
        return UnverifiedConceptualMetadata(
            space=self.data_model_id.space,
            external_id=self.data_model_id.external_id,
            # Validated in the constructor
            version=cast(str, self.data_model_id.version),
            name=name,
            creator="NEAT",
            created=now,
            updated=now,
            description="Inferred model from knowledge graph",
        )

    @classmethod
    def _add_uri_namespace_to_prefixes(cls, namespace: str, prefixes: dict[str, Namespace]) -> None:
        """Add URI to prefixes dict if not already present

        Args:
            URI: URI from which namespace is being extracted
            prefixes: Dict of prefixes and namespaces
        """
        if Namespace(namespace) not in prefixes.values():
            prefixes[f"prefix_{len(prefixes) + 1}"] = Namespace(namespace)



import uuid
from typing import Any

from cognite.neat._data_model.deployer.data_classes import DeploymentResult

from ._changes import SerializedChanges
from ._statistics import DeploymentStatistics


def serialize_deployment_result(result: DeploymentResult) -> dict[str, Any]:
    """Serialize deployment result into structured changes.

    Args:
        result: The deployment result to serialize.

    Returns:
        Serialized changes representing the deployment result.
    """
    result_dict = {"unique_id": uuid.uuid4().hex[:8], "status": result.status, "is_dry_run": result.is_dry_run}

    stats = DeploymentStatistics.from_deployment_result(result)
    changes = SerializedChanges.from_deployment_result(result)

    result_dict["total_changes"] = stats.total_changes
    result_dict["created"] = stats.by_change_type.create
    result_dict["updated"] = stats.by_change_type.update
    result_dict["deleted"] = stats.by_change_type.delete
    result_dict["skipped"] = stats.by_change_type.skip
    result_dict["unchanged"] = stats.by_change_type.unchanged
    result_dict["failed"] = stats.by_change_type.failed

    result_dict["STATS_JSON"] = stats.model_dump_json()
    result_dict["CHANGES_JSON"] = changes.model_dump_json_flat()

    return result_dict



from abc import ABC, abstractmethod
from typing import Any


class State(ABC):
    def __init__(self) -> None:
        # this will be reference to the actual store in the session
        # used to store data models and instances, here only as a placeholder
        self._store = None

    @abstractmethod
    def transition(self, event: Any) -> "State":
        """
        Handle events that are delegated to this State.
        """
        raise NotImplementedError("on_event() must be implemented by the subclass.")

    def can_transition(self, event: Any) -> bool:
        """
        Check if the state can transition on the given event.
        """
        # avoiding circular import
        from cognite.neat._state_machine._states import ForbiddenState

        return not isinstance(self.transition(event), ForbiddenState)

    def __repr__(self) -> str:
        """
        Leverages the __str__ method to describe the State.
        """
        return self.__str__()

    def __str__(self) -> str:
        """
        Returns the name of the State.
        """
        return self.__class__.__name__



from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.config import ClientConfig, global_config

from cognite.neat import _version
from cognite.neat._utils.repo import get_repo_root

from .credentials import get_credentials
from .env_vars import ClientEnvironmentVariables, parse_env_file
from .interactive import get_interactive_flow

CLIENT_NAME = f"CogniteNeat:{_version.__version__}"


def get_cognite_client(env_file_name: str) -> CogniteClient | None:
    """Get a CogniteClient using environment variables from a .env file."

    Args:
        env_file_name: The name of the .env file to look for in the repository root / current working directory. If
        the file is found, the variables will be loaded from the file. If the file is not found, the user will
        be prompted to enter the variables and the file will be created.

    Returns:
        CogniteClient: An instance of CogniteClient configured with the loaded environment variables.
    """
    # This function raises exceptions on failure
    if not env_file_name.endswith(".env"):
        raise ValueError(f"env_file_name must end with '.env'. Got: {env_file_name!r}")
    global_config.disable_pypi_version_check = True
    global_config.silence_feature_preview_warnings = True

    repo_root = get_repo_root()
    if repo_root and (env_path := repo_root / env_file_name).exists():
        print(f"Found {env_file_name} in repository root.")
    elif (env_path := Path.cwd() / env_file_name).exists():
        print(f"Found {env_file_name} in current working directory.")

    if env_path.exists():
        env_vars = parse_env_file(env_path)
        client_config = create_client_config_from_env_vars(env_vars)
        return CogniteClient(client_config)
    print(f"Failed to find {env_file_name} in repository root or current working directory.")

    env_folder = repo_root if repo_root is not None else Path.cwd()
    new_env_path = env_folder / env_file_name
    flow = get_interactive_flow(new_env_path)
    flow.run()
    print("Could not create CogniteClient because no environment file was found.")
    return None


def create_client_config_from_env_vars(env_vars: ClientEnvironmentVariables) -> ClientConfig:
    return ClientConfig(
        client_name=CLIENT_NAME,
        project=env_vars.CDF_PROJECT,
        credentials=get_credentials(env_vars),
        max_workers=env_vars.CDF_MAX_WORKERS,
        timeout=env_vars.CDF_CLIENT_TIMEOUT,
        base_url=env_vars.cdf_url,
    )



from abc import ABC
from typing import Annotated, Any, Literal

from pydantic import Field, Json, TypeAdapter, field_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from cognite.neat._utils.useful_types import BaseModelObject

from ._base import Resource, WriteableResource
from ._constants import CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN
from ._data_types import DataType
from ._references import ContainerDirectReference, ContainerReference, NodeReference, ViewDirectReference, ViewReference


class ViewPropertyDefinition(Resource, ABC):
    connection_type: str


class ViewCoreProperty(ViewPropertyDefinition, ABC):
    # Core properties do not have connection type in the API, but we add it here such that
    # we can use it as a discriminator in unions. The exclude=True ensures that it is not
    # sent to the API.
    connection_type: Literal["primary_property"] = Field(default="primary_property", exclude=True)
    name: str | None = Field(
        default=None,
        description="Readable property name.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the content and suggested use for this property..",
        max_length=1024,
    )
    container: ContainerReference = Field(
        description="Reference to the container where this property is defined.",
    )
    container_property_identifier: str = Field(
        description="Identifier of the property in the container.",
        min_length=1,
        max_length=255,
        pattern=CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    )

    @field_serializer("container", mode="plain")
    @classmethod
    def serialize_container(cls, container: ContainerReference, info: FieldSerializationInfo) -> dict[str, Any]:
        output = container.model_dump(**vars(info))
        output["type"] = "container"
        return output


class ViewCorePropertyRequest(ViewCoreProperty):
    source: ViewReference | None = Field(
        default=None,
        description="Indicates on what type a referenced direct relation is expected to be. "
        "Only applicable for direct relation properties.",
    )

    @field_serializer("source", mode="plain")
    @classmethod
    def serialize_source(cls, source: ViewReference | None, info: FieldSerializationInfo) -> dict[str, Any] | None:
        if source is None:
            return None
        output = source.model_dump(**vars(info))
        output["type"] = "view"
        return output


class ConstraintOrIndexState(BaseModelObject):
    nullability: Literal["current", "pending", "failed"] | None = Field(
        None,
        description="""For properties that have isNullable set to false, this field describes the validity of the
not-null constraint. It is not specified for nullable properties.

Possible values are:

"failed": The property contains null values, violating the constraint. This can occur if a property with
          existing nulls was made non-nullable. New null values will still be rejected.
"current": The constraint is satisfied; all values in the property are not null.
"pending": The constraint validity has not yet been computed.
        """,
    )


class ViewCorePropertyResponse(ViewCoreProperty, WriteableResource[ViewCorePropertyRequest]):
    immutable: bool | None = Field(
        default=None,
        description="Should updates to this property be rejected after the initial population?",
    )
    nullable: bool | None = Field(
        default=None,
        description="Does this property need to be set to a value, or not?",
    )
    auto_increment: bool | None = Field(
        default=None,
        description="Increment the property based on its highest current value (max value).",
    )
    default_value: str | int | bool | dict[str, Json] | None = Field(
        default=None,
        description="Default value to use when you do not specify a value for the property.",
    )
    constraint_state: ConstraintOrIndexState = Field(
        description="Describes the validity of constraints defined on this property"
    )
    type: DataType = Field(description="The type of data you can store in this property.")

    def as_request(self) -> ViewCorePropertyRequest:
        return ViewCorePropertyRequest.model_validate(self.model_dump(by_alias=True))


class ConnectionPropertyDefinition(ViewPropertyDefinition, ABC):
    name: str | None = Field(
        default=None,
        description="Readable property name.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the content and suggested use for this property..",
        max_length=1024,
    )


class EdgeProperty(ConnectionPropertyDefinition, ABC):
    source: ViewReference = Field(
        description="The target node(s) of this connection can be read through the view specified in 'source'."
    )
    type: NodeReference = Field(
        description="Reference to the node pointed to by the direct relation. The reference consists of a "
        "space and an external-id."
    )
    edge_source: ViewReference | None = Field(
        None, description="The edge(s) of this connection can be read through the view specified in 'edgeSource'."
    )
    direction: Literal["outwards", "inwards"] = Field(
        "outwards", description="The direction of the edge(s) of this connection."
    )

    @field_serializer("source", "edge_source", mode="plain")
    @classmethod
    def serialize_source(cls, source: ViewReference | None, info: FieldSerializationInfo) -> dict[str, Any] | None:
        if source is None:
            return None
        output = source.model_dump(**vars(info))
        output["type"] = "view"
        return output


class SingleEdgeProperty(EdgeProperty):
    connection_type: Literal["single_edge_connection"] = "single_edge_connection"


class MultiEdgeProperty(EdgeProperty):
    connection_type: Literal["multi_edge_connection"] = "multi_edge_connection"


class ReverseDirectRelationProperty(ConnectionPropertyDefinition, ABC):
    source: ViewReference = Field(
        description="The node(s) containing the direct relation property can be read "
        "through the view specified in 'source'."
    )
    through: ContainerDirectReference | ViewDirectReference = Field(
        description="The view of the node containing the direct relation property."
    )

    @field_serializer("source", mode="plain")
    @classmethod
    def serialize_source(cls, source: ViewReference, info: FieldSerializationInfo) -> dict[str, Any] | None:
        output = source.model_dump(**vars(info))
        output["type"] = "view"
        return output

    @field_serializer("through", mode="plain")
    @classmethod
    def serialize_through(
        cls, through: ContainerDirectReference | ViewDirectReference, info: FieldSerializationInfo
    ) -> dict[str, Any]:
        output = through.model_dump(**vars(info))
        if isinstance(through, ContainerDirectReference):
            output["source"]["type"] = "container"
        else:
            output["source"]["type"] = "view"
        return output


class SingleReverseDirectRelationPropertyRequest(ReverseDirectRelationProperty):
    connection_type: Literal["single_reverse_direct_relation"] = "single_reverse_direct_relation"


class MultiReverseDirectRelationPropertyRequest(ReverseDirectRelationProperty):
    connection_type: Literal["multi_reverse_direct_relation"] = "multi_reverse_direct_relation"


class SingleReverseDirectRelationPropertyResponse(
    ReverseDirectRelationProperty, WriteableResource[SingleReverseDirectRelationPropertyRequest]
):
    connection_type: Literal["single_reverse_direct_relation"] = "single_reverse_direct_relation"
    targets_list: bool = Field(
        description="Whether or not this reverse direct relation targets a list of direct relations.",
    )

    def as_request(self) -> SingleReverseDirectRelationPropertyRequest:
        return SingleReverseDirectRelationPropertyRequest.model_validate(self.model_dump(by_alias=True))


class MultiReverseDirectRelationPropertyResponse(
    ReverseDirectRelationProperty, WriteableResource[MultiReverseDirectRelationPropertyRequest]
):
    connection_type: Literal["multi_reverse_direct_relation"] = "multi_reverse_direct_relation"
    targets_list: bool = Field(
        description="Whether or not this reverse direct relation targets a list of direct relations.",
    )

    def as_request(self) -> MultiReverseDirectRelationPropertyRequest:
        return MultiReverseDirectRelationPropertyRequest.model_validate(self.model_dump(by_alias=True))


ViewRequestProperty = Annotated[
    SingleEdgeProperty
    | MultiEdgeProperty
    | SingleReverseDirectRelationPropertyRequest
    | MultiReverseDirectRelationPropertyRequest
    | ViewCorePropertyRequest,
    Field(discriminator="connection_type"),
]
ViewResponseProperty = Annotated[
    SingleEdgeProperty
    | MultiEdgeProperty
    | SingleReverseDirectRelationPropertyResponse
    | MultiReverseDirectRelationPropertyResponse
    | ViewCorePropertyResponse,
    Field(discriminator="connection_type"),
]

ViewRequestPropertyAdapter: TypeAdapter[ViewRequestProperty] = TypeAdapter(ViewRequestProperty)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Generic, TypeVar

from cognite.neat._data_model.models.dms import RequestSchema

T_Export = TypeVar("T_Export")


class DMSExporter(ABC, Generic[T_Export]):
    """This is the base class for all DMS exporters."""

    NEW_LINE = "\n"
    ENCODING = "utf-8"

    @abstractmethod
    def export(self, data_model: RequestSchema) -> T_Export:
        raise NotImplementedError()


class DMSFileExporter(DMSExporter[T_Export], ABC):
    @abstractmethod
    def export_to_file(self, data_model: RequestSchema, file_path: Path) -> None:
        raise NotImplementedError()



from cognite.neat._data_model._constants import COGNITE_APP_SPACES, COGNITE_SPACES
from cognite.neat._data_model.rules.cdf._base import CDFRule
from cognite.neat._issues import Recommendation

BASE_CODE = "NEAT-CDF-SPACES"


class EmptySpaces(CDFRule):
    """Rule that checks for empty spaces in CDF.

    ## What it does
    This rule checks if there are any empty spaces in CDF.

    ## Why is this bad?
    CDF projects typically have limits of 100 spaces, and having empty spaces can waste these valuable resources.
    Also, empty spaces can lead to confusion and mismanagement of resources within the CDF environment.
    They may indicate incomplete configurations or unused resources that could be cleaned up.

    ## Example
    A space `iamempty` with no associated resources such as Views, Containers or Data Models.


    """

    code = f"{BASE_CODE}-001"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        issues: list[Recommendation] = []

        if not self.validation_resources.space_statistics:
            return issues

        empty_spaces = set(self.validation_resources.space_statistics.empty_spaces()) - set(
            COGNITE_APP_SPACES + COGNITE_SPACES
        )

        for space in empty_spaces:
            issues.append(
                Recommendation(
                    message=f"Space '{space}' is empty and has no associated resources.",
                    code=self.code,
                    fix="Consider removing the empty space to maintain a clean CDF environment.",
                )
            )

        return issues



"""Plugin manager for external plugins."""

from importlib import metadata
from typing import Any, ClassVar

from ._base import NeatPlugin
from ._data_model import DataModelImporterPlugin, DataModelTransformerPlugin
from ._issues import PluginDuplicateError, PluginError, PluginLoadingError


class Plugin:
    """Plugin class for registering plugins registered via entry points (i.e. external plugins).

    Args:
        name (str): The name of format (e.g. Excel) or action (e.g. merge) plugin is handling.
        type_ (type): The type of the plugin.
        entry_point (EntryPoint): The entry point for the plugin.

    !!! note "name uniqueness"
        The name of the plugin must be lower case and unique across all plugins of the same kind.
        If two plugins have the same name, the exception will be raised.
    """

    def __init__(self, name: str, type_: type[NeatPlugin], entry_point: metadata.EntryPoint):
        self.name = name
        self.type_ = type_
        self.entry_point = entry_point

    def load(self) -> type[NeatPlugin]:
        try:
            return self.entry_point.load()
        except Exception as e:
            raise PluginLoadingError(self.name, self.type_.__name__, str(e)) from e


class PluginManager:
    """Plugin manager for external plugins."""

    _plugins_entry_points: ClassVar[dict[str, type[NeatPlugin]]] = {
        "cognite.neat._v0.plugins.data_model.importers": DataModelImporterPlugin,
        "cognite.neat._v0.plugins.data_model.transformers": DataModelTransformerPlugin,
    }

    def __init__(self, plugins: dict[tuple[str, type[NeatPlugin]], Any]) -> None:
        self._plugins = plugins

    def get(self, name: str, type_: type[NeatPlugin]) -> type[NeatPlugin]:
        """
        Returns desired plugin

        Args:
            name (str): The name of format (e.g. Excel) or action (e.g. merge) plugin is handling.
            type_ (type): The type of the plugin.
        """
        try:
            plugin_class = self._plugins[(name, type_)]
            return plugin_class
        except KeyError:
            raise PluginError(plugin_name=name, plugin_type=type_.__name__) from None

    @classmethod
    def load_plugins(cls, entry_points: metadata.EntryPoints | None = None) -> "PluginManager":
        """Load plugins from entry points and register them to the manager.
        This method scans the provided entry points for plugins defined in the
        `_plugins_entry_points` dictionary and registers them.

        Args:
            entry_points: Entry points to load plugins from. If None, uses the default entry points.
        """
        _plugins: dict[tuple[str, type[NeatPlugin]], Any] = {}

        entry_points = entry_points or metadata.entry_points()
        if hasattr(entry_points, "select"):
            for group, type_ in cls._plugins_entry_points.items():
                for entry_point in entry_points.select(group=group):
                    # Check for duplicate plugins
                    if (entry_point.name, type_) in _plugins:
                        raise PluginDuplicateError(plugin_name=entry_point.name, plugin_type=type_.__name__)

                    # Register the plugin
                    _plugins[(entry_point.name, type_)] = Plugin(entry_point.name, type_, entry_point).load()

        return cls(_plugins)


_manager_instance: PluginManager | None = None


def get_plugin_manager(force_reload: bool = False) -> PluginManager:
    """Get or create a singleton PluginManager instance."""
    global _manager_instance
    if force_reload or _manager_instance is None:
        _manager_instance = PluginManager.load_plugins()
    return _manager_instance



import subprocess
from pathlib import Path


def get_repo_root() -> Path | None:
    """Get the root path of the git repository.

    Raises:
        RuntimeError: If git is not installed or the current directory is not in a git repository

    """
    try:
        result = subprocess.run("git rev-parse --show-toplevel".split(), stdout=subprocess.PIPE)
    except FileNotFoundError:
        # Git is not installed or not found in PATH
        return None
    output = result.stdout.decode().strip()
    if not output:
        # Not in a git repository
        return None
    return Path(output)



import itertools
from typing import Any

from pydantic import BaseModel, Field

from cognite.neat._data_model.deployer.data_classes import (
    AddedField,
    AppliedChanges,
    ChangedField,
    ChangeResult,
    DeploymentResult,
    FieldChange,
    FieldChanges,
    RemovedField,
    ResourceChange,
)


class SerializedFieldChange(BaseModel):
    """Serialized field change for JSON output."""

    field_path: str
    severity: str
    description: str

    @classmethod
    def from_field_change(cls, field_change: FieldChange) -> list["SerializedFieldChange"]:
        """Serialize a field change, handling nested FieldChanges recursively.

        Args:
            field_change: The field change to serialize.

        Returns:
            List of serialized field changes (may be multiple if nested).
        """
        serialized_changes: list[SerializedFieldChange] = []

        if isinstance(field_change, FieldChanges):
            # Recursively handle nested changes
            for change in field_change.changes:
                serialized_changes.extend(cls.from_field_change(change))
        else:
            # Base case: single field change
            serialized_changes.append(cls._from_single_field_change(field_change))

        return serialized_changes

    @classmethod
    def _from_single_field_change(cls, field_change: FieldChange) -> "SerializedFieldChange":
        """Serialize a single non-nested field change.

        Args:
            field_change: The single field change to serialize.

        Returns:
            Serialized field change.
        """
        return cls(
            field_path=field_change.field_path,
            severity=field_change.severity.name,
            description=field_change.description
            if isinstance(field_change, AddedField | RemovedField | ChangedField)
            else "Field changed",
        )


class SerializedResourceChange(BaseModel):
    """Serialized resource change for JSON output."""

    id: int
    endpoint: str
    change_type: str
    severity: str
    resource_id: str
    message: str | None = None
    changes: list[SerializedFieldChange] = Field(default_factory=list)

    @classmethod
    def from_resource_change(
        cls, resource: ResourceChange, endpoint: str, change_id: int
    ) -> "SerializedResourceChange":
        """Serialize a single resource change.

        Args:
            resource: The resource change to serialize.
            endpoint: The endpoint type.
            change_id: Unique ID for this change.

        Returns:
            Serialized resource change.
        """
        changes: list[SerializedFieldChange] = []

        for change in resource.changes:
            changes.extend(SerializedFieldChange.from_field_change(change))

        return cls(
            id=change_id,
            endpoint=endpoint,
            change_type=resource.change_type,
            severity=resource.severity.name,
            resource_id=str(resource.resource_id),
            changes=changes,
            message=resource.message,
        )

    @classmethod
    def from_change_result(cls, change_id: int, response: ChangeResult) -> "SerializedResourceChange":
        """Serialize from a change result (actual deployment).

        Args:
            change_id: Unique ID for this change.
            response: The change result from deployment.

        Returns:
            Serialized resource change with deployment status.
        """
        serialized_resource_change = cls.from_resource_change(
            resource=response.change,
            endpoint=response.endpoint,
            change_id=change_id,
        )

        serialized_resource_change.message = response.message
        if not response.is_success:
            serialized_resource_change.change_type = "failed"

        return serialized_resource_change


class SerializedChanges(BaseModel):
    """Container for all serialized changes."""

    changes: list[SerializedResourceChange] = Field(default_factory=list)

    @classmethod
    def from_deployment_result(cls, result: DeploymentResult) -> "SerializedChanges":
        """Create SerializedChanges from a DeploymentResult.

        Args:
            result: The deployment result to serialize changes from.

        Returns:
            SerializedChanges instance with all changes.
        """
        serialized = cls()

        if not result.responses:
            serialized._add_from_dry_run(result)
        else:
            serialized._add_from_applied_changes(result.responses)

        return serialized

    def _add_from_dry_run(self, result: DeploymentResult) -> None:
        """Add changes from dry run deployment.

        Args:
            result: The deployment result in dry run mode.
        """
        # Iterate over each endpoint plan
        for endpoint_plan in result.plan:
            # Then per resource in the endpoint
            for resource in endpoint_plan.resources:
                # Then serialize individual resource change
                serialized_resource_change = SerializedResourceChange.from_resource_change(
                    resource=resource,
                    endpoint=endpoint_plan.endpoint,
                    change_id=len(self.changes),
                )
                self.changes.append(serialized_resource_change)

    def _add_from_applied_changes(self, applied_changes: AppliedChanges) -> None:
        """Add changes from actual deployment.
        Args:
            result: The deployment result from actual deployment.
        """
        for response in itertools.chain(
            applied_changes.created,
            applied_changes.merged_updated,
            applied_changes.deletions,
            applied_changes.unchanged,
            applied_changes.skipped,
        ):
            self.changes.append(SerializedResourceChange.from_change_result(len(self.changes), response))

    def model_dump_json_flat(self, **kwargs: Any) -> str:
        """Dump changes as JSON array without the wrapper key.
        Returns:
            JSON string of the changes array.
        """
        if not self.changes:
            return "[]"

        iterator = (change.model_dump_json(**kwargs) for change in self.changes)
        return f"[{','.join(iterator)}]"



import importlib

from cognite.neat._exceptions import NeatImportError

try:
    importlib.import_module("rdflib")
except ImportError as e:
    raise NeatImportError("legacy module", "legacy") from e


from cognite.neat._v0.core._utils.auth import get_cognite_client
from cognite.neat._v0.session._base import NeatSession

from ._version import __version__

__all__ = ["NeatSession", "__version__", "get_cognite_client"]



"""This module performs importing of various formats to one of serializations for which
there are loaders to data model pydantic class."""

from cognite.neat._v0.core._data_model.importers._rdf._base import BaseRDFImporter
from cognite.neat._v0.core._data_model.importers._rdf._shared import (
    parse_concepts,
    parse_properties,
)

CLASSES_QUERY = """SELECT ?concept  ?name ?description ?implements
        WHERE {{

        ?concept  a owl:Class .
        OPTIONAL {{?concept  rdfs:subClassOf ?implements }}.
        OPTIONAL {{?concept  rdfs:label|skos:prefLabel ?name }}.
        OPTIONAL {{?concept  rdfs:comment|skos:definition ?description}} .


        FILTER (!isBlank(?concept ))
        FILTER (!bound(?implements) || !isBlank(?implements))

        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "{language}"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "{language}"))

    }}
    """

CLASSES_QUERY_PARAMETERS = {"concept", "name", "description", "implements"}

PROPERTIES_QUERY = """ SELECT ?concept ?property_ ?name ?description ?value_type ?min_count ?max_count ?default
    WHERE {{
        ?property_ a ?property_Type.
        FILTER (?property_Type IN (owl:ObjectProperty, owl:DatatypeProperty ) )

        # --- 1. Explicit Domain Discovery ---

        # A. Handling owl:domain when it is expressed as owl restriction
        OPTIONAL {{
            ?property_ rdfs:domain ?domain_exp_node .
            FILTER(isBlank(?domain_exp_node))
            ?domain_exp_node owl:unionOf|owl:intersectionOf ?exp_concepts_list .
            ?exp_concepts_list rdf:rest*/rdf:first ?explicit_concept.
        }}

        # B. Handling the domain when it is a single concept
        OPTIONAL {{
            ?property_ rdfs:domain ?domain_exp_node .
            FILTER(!isBlank(?domain_exp_node))
            BIND(?domain_exp_node AS ?explicit_concept)
        }}

        # --- 2. Inherited Domain Discovery (Fallback) ---

        # C. Handling inherited domain when parent domain is a restriction
        OPTIONAL {{
            ?property_ rdfs:subPropertyOf ?parent_property .
            ?parent_property rdfs:domain ?parent_domain_node .
            FILTER(isBlank(?parent_domain_node))
            ?parent_domain_node owl:unionOf|owl:intersectionOf ?parent_concepts_list .
            ?parent_concepts_list rdf:rest*/rdf:first ?inherited_concept.
        }}

        # D. Handling inherited domain when parent domain is a single concept
        OPTIONAL {{
            ?property_ rdfs:subPropertyOf ?parent_property .
            ?parent_property rdfs:domain ?parent_domain_node .
            FILTER(!isBlank(?parent_domain_node))
            BIND(?parent_domain_node AS ?inherited_concept)
        }}

        # Final Concept Assignment with Priority ---
        # COALESCE prioritizes ?explicit_concept over ?inherited_concept
        BIND(COALESCE(?explicit_concept, ?inherited_concept) AS ?concept)


        # Handling owl:range when it is expressed as owl restriction
        OPTIONAL {{
            ?property_ rdfs:range ?range .
            FILTER(isBlank(?range))
            ?range owl:unionOf|owl:intersectionOf ?value_types .
            ?value_types rdf:rest*/rdf:first ?value_type.
        }}

        # Handling the range when it is a single concept
        OPTIONAL {{
            ?property_ rdfs:range ?range .
            FILTER(!isBlank(?range))
            BIND(?range AS ?value_type)
        }}

        OPTIONAL {{?property_ rdfs:label|skos:prefLabel ?name }}.
        OPTIONAL {{?property_ rdfs:comment|skos:definition ?description}}.
        OPTIONAL {{?property_ owl:maxCardinality ?max_count}}.
        OPTIONAL {{?property_ owl:minCardinality ?min_count}}.

        # FILTERS
        FILTER (!isBlank(?property_))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "{language}"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "{language}"))
    }}
    """
PROPERTIES_QUERY_PARAMETERS = {
    "concept",
    "property_",
    "name",
    "description",
    "value_type",
    "min_count",
    "max_count",
    "default",
}


class OWLImporter(BaseRDFImporter):
    """Convert OWL ontology to unverified data model.

    Args:
        filepath: Path to OWL ontology
    """

    def _to_data_model_components(
        self,
    ) -> dict:
        concepts, issue_list = parse_concepts(
            self.graph, CLASSES_QUERY, CLASSES_QUERY_PARAMETERS, self.language, self.issue_list
        )
        self.issue_list = issue_list

        properties, issue_list = parse_properties(
            self.graph, PROPERTIES_QUERY, PROPERTIES_QUERY_PARAMETERS, self.language, self.issue_list
        )
        self.issue_list = issue_list

        components = {
            "Metadata": self._metadata,
            "Concepts": list(concepts.values()) if concepts else [],
            "Properties": list(properties.values()) if properties else [],
        }

        return components

    @property
    def description(self) -> str:
        return f"Ontology {self.source_name} read as unverified data model"



from collections.abc import Callable

from cognite.neat._client.data_classes import SpaceStatisticsResponse
from cognite.neat._data_model._analysis import ValidationResources
from cognite.neat._data_model._shared import OnSuccessIssuesChecker
from cognite.neat._data_model._snapshot import SchemaSnapshot
from cognite.neat._data_model.models.dms._limits import SchemaLimits
from cognite.neat._utils.auxiliary import get_concrete_subclasses

from ._base import CDFRule


class CDFRulesOrchestrator(OnSuccessIssuesChecker):
    """CDF rules orchestrator, used to execute CDF rules on an entire CDF snapshot."""

    def __init__(
        self,
        limits: SchemaLimits,
        space_statistics: SpaceStatisticsResponse,
        can_run_validator: Callable[[str, type], bool] | None = None,
        enable_alpha_validators: bool = False,
    ) -> None:
        super().__init__()
        self._limits = limits
        self._can_run_validator = can_run_validator or (lambda code, issue_type: True)  # type: ignore
        self._has_run = False
        self._enable_alpha_validators = enable_alpha_validators
        self._space_statistics = space_statistics

    def run(self, cdf_snapshot: SchemaSnapshot) -> None:
        """Run quality assessment on the DMS data model."""

        validation_resources = self._gather_validation_resources(cdf_snapshot)

        # Initialize all validators
        validators: list[CDFRule] = [validator(validation_resources) for validator in get_concrete_subclasses(CDFRule)]

        # Run validators
        for validator in validators:
            if validator.alpha and not self._enable_alpha_validators:
                continue
            if self._can_run_validator(validator.code, validator.issue_type):
                self._issues.extend(validator.validate())

        self._has_run = True

    def _gather_validation_resources(self, cdf_snapshot: SchemaSnapshot) -> ValidationResources:
        # we do not want to modify the original request schema during validation

        return ValidationResources(
            cdf=cdf_snapshot,
            local=cdf_snapshot,
            limits=self._limits,
            space_statistics=self._space_statistics,
            modus_operandi="rebuild",
        )



from pathlib import Path

shared_style = Path(__file__).parent / "shared.css"
issues_style = Path(__file__).parent / "issues.css"
issues_scripts = Path(__file__).parent / "issues.js"

deployment_style = Path(__file__).parent / "deployment.css"
deployment_scripts = Path(__file__).parent / "deployment.js"

statistics_style = Path(__file__).parent / "statistics.css"
statistics_scripts = Path(__file__).parent / "statistics.js"



from ._api_importer import DMSAPICreator, DMSAPIImporter
from ._base import DMSImporter
from ._table_importer.importer import DMSTableImporter

__all__ = ["DMSAPICreator", "DMSAPIImporter", "DMSImporter", "DMSTableImporter"]



import math
import warnings
from collections.abc import Hashable
from typing import TYPE_CHECKING, Any, ClassVar, Literal

import pandas as pd
from cognite.client import data_modeling as dm
from pydantic import Field, field_serializer, field_validator, model_validator
from pydantic_core.core_schema import SerializationInfo, ValidationInfo

from cognite.neat._v0.core._client.data_classes.schema import DMSSchema
from cognite.neat._v0.core._constants import DMS_CONTAINER_LIST_MAX_LIMIT
from cognite.neat._v0.core._data_model._constants import CONSTRAINT_ID_MAX_LENGTH
from cognite.neat._v0.core._data_model.models._base_verified import (
    BaseVerifiedDataModel,
    BaseVerifiedMetadata,
    ContainerProperty,
    DataModelLevel,
    RoleTypes,
    SheetList,
    SheetRow,
    ViewProperty,
    ViewRef,
)
from cognite.neat._v0.core._data_model.models._types import (
    ConceptEntityType,
    ContainerEntityType,
    PhysicalPropertyType,
    URIRefType,
    ViewEntityType,
)
from cognite.neat._v0.core._data_model.models.data_types import DataType
from cognite.neat._v0.core._data_model.models.entities import (
    ConceptualEntity,
    ContainerIndexEntity,
    DMSNodeEntity,
    EdgeEntity,
    HasDataFilter,
    NodeTypeFilter,
    PhysicalEntity,
    PhysicalUnknownEntity,
    RawFilter,
    ReverseConnectionEntity,
    Undefined,
    ViewEntity,
    ViewEntityList,
)
from cognite.neat._v0.core._data_model.models.entities._types import (
    ContainerConstraintListType,
    ContainerIndexListType,
)
from cognite.neat._v0.core._issues.errors import NeatValueError
from cognite.neat._v0.core._issues.warnings import NeatValueWarning, PropertyDefinitionWarning

if TYPE_CHECKING:
    from cognite.neat._v0.core._data_model.models import ConceptualDataModel

_DEFAULT_VERSION = "1"


class PhysicalMetadata(BaseVerifiedMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.dms
    level: ClassVar[DataModelLevel] = DataModelLevel.physical
    conceptual: URIRefType | None = None

    def as_space(self) -> dm.SpaceApply:
        return dm.SpaceApply(
            space=self.space,
        )

    def as_data_model(self) -> dm.DataModelApply:
        suffix = f"Creator: {', '.join(self.creator)}"
        if self.description:
            description = f"{self.description} Creator: {', '.join(self.creator)}"
        else:
            description = suffix

        return dm.DataModelApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name or None,
            version=self.version or "missing",
            description=description,
            views=[],
        )

    def get_prefix(self) -> str:
        return self.space


def _metadata(context: Any) -> PhysicalMetadata | None:
    if isinstance(context, dict) and isinstance(context.get("metadata"), PhysicalMetadata):
        return context["metadata"]
    return None


class PhysicalProperty(SheetRow):
    """Physical property provides a complete definition of a property in the data model in a physical data model.
    This includes view to which the property belongs as well mapping between the view and container property.
    """

    view: ViewEntityType = Field(alias="View", description="The property identifier.")
    view_property: PhysicalPropertyType = Field(
        alias="View Property", description="The ViewId this property belongs to"
    )
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the property")
    description: str | None = Field(alias="Description", default=None, description="Short description of the property")
    connection: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None = Field(
        None,
        alias="Connection",
        description="nly applies to connection between views. "
        "It specify how the connection should be implemented in CDF.",
    )
    value_type: DataType | ViewEntity | PhysicalUnknownEntity = Field(
        alias="Value Type",
        description="Value type that the property can hold. It takes either subset of CDF primitive types or a View id",
    )
    min_count: int | None = Field(
        alias="Min Count",
        default=None,
        description="Minimum number of values that the property can hold. "
        "If no value is provided, the default value is  `0`, "
        "which means that the property is optional.",
    )
    max_count: int | float | None = Field(
        alias="Max Count",
        default=None,
        description="Maximum number of values that the property can hold. "
        "If no value is provided, the default value is  `inf`, "
        "which means that the property can hold any number of values (listable).",
    )
    immutable: bool | None = Field(
        default=None,
        alias="Immutable",
        description="sed to indicate whether the property is can only be set once. Only applies to primitive type.",
    )
    default: bool | str | int | float | dict | None = Field(
        None, alias="Default", description="Specifies default value for the property."
    )
    container: ContainerEntityType | None = Field(
        None,
        alias="Container",
        description="Specifies container where the property is stored. Only applies to primitive type.",
    )
    container_property: PhysicalPropertyType | None = Field(
        None,
        alias="Container Property",
        description="Specifies property in the container where the property is stored. Only applies to primitive type.",
    )
    index: ContainerIndexListType | None = Field(
        None,
        alias="Index",
        description="The names of the indexes (comma separated) that should be created for the property.",
    )
    constraint: ContainerConstraintListType | None = Field(
        None,
        alias="Constraint",
        description="The names of the uniquness (comma separated) that should be created for the property.",
    )
    conceptual: URIRefType | None = Field(
        None,
        alias="Conceptual",
        description="Used to make connection between physical and conceptual data model aspect",
    )

    @property
    def nullable(self) -> bool | None:
        """Used to indicate whether the property is required or not. Only applies to primitive type."""
        return self.min_count in {0, None}

    @property
    def is_list(self) -> bool | None:
        """Used to indicate whether the property holds single or multiple values (list). "
        "Only applies to primitive types."""
        if self.max_count is None:
            return None
        return self.max_count is float("inf") or (isinstance(self.max_count, int | float) and self.max_count > 1)

    def _identifier(self) -> tuple[Hashable, ...]:
        return self.view, self.view_property

    @field_validator("min_count")
    def direct_relation_must_be_nullable(cls, value: Any, info: ValidationInfo) -> None:
        if info.data.get("connection") == "direct" and value not in {0, None}:
            raise ValueError("Direct relation must have min count set to 0")
        return value

    @field_validator("max_count", mode="before")
    def as_integer(cls, value: Any) -> Any:
        if isinstance(value, float) and not math.isinf(value):
            return int(value)
        return value

    @field_validator("max_count")
    def max_list_size(cls, value: Any, info: ValidationInfo) -> Any:
        if isinstance(info.data.get("connection"), EdgeEntity | ReverseConnectionEntity):
            if value is not None and value != float("inf") and not (isinstance(value, int) and value == 1):
                raise ValueError("Edge and reverse connections must have max count set to inf or 1")
            return value
        # We do not have a connection, so we can check the max list size.
        if isinstance(value, int) and value > DMS_CONTAINER_LIST_MAX_LIMIT:
            raise ValueError(f"Max list size cannot be greater than {DMS_CONTAINER_LIST_MAX_LIMIT}")
        return value

    @field_validator("value_type", mode="after")
    def connections_value_type(
        cls, value: EdgeEntity | ViewEntity | PhysicalUnknownEntity, info: ValidationInfo
    ) -> DataType | EdgeEntity | ViewEntity | PhysicalUnknownEntity:
        if (connection := info.data.get("connection")) is None:
            if isinstance(value, ViewEntity):
                raise ValueError(
                    f"Missing connection type for property '{info.data.get('view_property', 'unknown')}'. This "
                    f"is required with value type pointing to another view."
                )
            return value
        if connection == "direct" and not isinstance(value, ViewEntity | PhysicalUnknownEntity):
            raise ValueError(f"Direct relation must have a value type that points to a view, got {value}")
        elif isinstance(connection, EdgeEntity) and not isinstance(value, ViewEntity):
            raise ValueError(f"Edge connection must have a value type that points to a view, got {value}")
        elif isinstance(connection, ReverseConnectionEntity) and not isinstance(value, ViewEntity):
            raise ValueError(f"Reverse connection must have a value type that points to a view, got {value}")
        return value

    @field_validator("default", mode="after")
    def set_proper_type_on_default(cls, value: Any, info: ValidationInfo) -> Any:
        if not value:
            return value
        value_type = info.data.get("value_type")
        if not isinstance(value_type, DataType):
            warnings.filterwarnings("default")
            warnings.warn(
                NeatValueWarning(f"Default value {value} set to connection {value_type} will be ignored"),
                stacklevel=2,
            )
            return None
        else:
            try:
                return value_type.convert_value(value)
            except ValueError:
                warnings.filterwarnings("default")
                warnings.warn(
                    NeatValueWarning(f"Could not convert {value} to {value_type}"),
                    stacklevel=2,
                )
                return None

    @field_validator("container", "container_property", mode="after")
    def container_set_correctly(cls, value: Any, info: ValidationInfo) -> Any:
        if (connection := info.data.get("connection")) is None:
            return value
        if connection == "direct" and value is None:
            raise ValueError(
                "You must provide a container and container property for where to store direct connections"
            )
        elif isinstance(connection, EdgeEntity) and value is not None:
            raise ValueError(
                "Edge connections are not stored in a container, please remove the container and container property"
            )
        elif isinstance(connection, ReverseConnectionEntity) and value is not None:
            raise ValueError(
                "Reverse connection are not stored in a container, please remove the container and container property"
            )
        return value

    @field_validator("index", mode="after")
    @classmethod
    def index_set_correctly(cls, value: list[ContainerIndexEntity] | None, info: ValidationInfo) -> Any:
        if value is None:
            return value

        container = info.data["container"]
        container_property = info.data["container_property"]

        if not container or not container_property:
            raise ValueError("Container and container property must be set to use indexes")
        max_count = info.data.get("max_count")
        is_list = (
            max_count is not None and (isinstance(max_count, int | float) and max_count > 1)
        ) or max_count is float("inf")
        for index in value:
            if index.prefix is Undefined:
                message = f"The type of index is not defined. Please set 'inverted:{index!s}' or 'btree:{index!s}'."
                warnings.warn(
                    PropertyDefinitionWarning(str(container), "container property", str(container_property), message),
                    stacklevel=2,
                )
            elif index.prefix == "inverted" and not is_list:
                message = (
                    "It is not recommended to use inverted index on non-list properties. "
                    "Please consider using btree index instead."
                )
                warnings.warn(
                    PropertyDefinitionWarning(str(container), "container property", str(container_property), message),
                    stacklevel=2,
                )
            elif index.prefix == "btree" and is_list:
                message = (
                    "It is not recommended to use btree index on list properties. "
                    "Please consider using inverted index instead."
                )
                warnings.warn(
                    PropertyDefinitionWarning(str(container), "container property", str(container_property), message),
                    stacklevel=2,
                )
            if index.prefix == "inverted" and (index.cursorable is not None or index.by_space is not None):
                message = "Cursorable and bySpace are not supported for inverted indexes. These will be ignored."
                warnings.warn(
                    PropertyDefinitionWarning(str(container), "container property", str(container_property), message),
                    stacklevel=2,
                )
        return value

    @field_validator("constraint", mode="after")
    @classmethod
    def constraint_set_correctly(cls, value: ContainerConstraintListType | None, info: ValidationInfo) -> Any:
        if value is None:
            return value

        container = info.data["container"]
        container_property = info.data["container_property"]

        if not container or not container_property:
            raise ValueError("Container and container property must be set to use constraint")

        for constraint in value:
            if constraint.prefix is Undefined:
                message = f"The type of constraint is not defined. Please set 'uniqueness:{constraint!s}'."
                warnings.warn(
                    PropertyDefinitionWarning(str(container), "container property", str(container_property), message),
                    stacklevel=2,
                )
            elif constraint.prefix != "uniqueness":
                message = (
                    f"Unsupported constraint type on container property"
                    f" '{constraint.prefix}'. Currently only 'uniqueness' is supported."
                )
                raise ValueError(message) from None

            if len(constraint.suffix) > CONSTRAINT_ID_MAX_LENGTH:
                message = f"Constraint id '{constraint.suffix}' exceeds maximum length of {CONSTRAINT_ID_MAX_LENGTH}."
                raise ValueError(message) from None

        return value

    @field_serializer("value_type", when_used="always")
    def as_dms_type(self, value_type: DataType | EdgeEntity | ViewEntity, info: SerializationInfo) -> str:
        if isinstance(value_type, DataType):
            return value_type._suffix_extra_args(value_type.dms._type)
        elif isinstance(value_type, EdgeEntity | ViewEntity) and (metadata := _metadata(info.context)):
            return value_type.dump(space=metadata.space, version=metadata.version)
        return str(value_type)

    @field_serializer("view", "container", when_used="unless-none")
    def remove_default_space(self, value: str, info: SerializationInfo) -> str:
        if (metadata := _metadata(info.context)) and isinstance(value, ConceptualEntity):
            if info.field_name == "container" and info.context.get("as_reference") is True:
                # When dumping as reference, the container should keep the default space for easy copying
                # over to user sheets.
                return value.dump()
            return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("connection", when_used="unless-none")
    def remove_defaults(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, ConceptualEntity) and (metadata := _metadata(info.context)):
            default_type = f"{self.view.external_id}.{self.view_property}"
            if isinstance(value, EdgeEntity) and value.edge_type and value.edge_type.space != metadata.space:
                default_type = f"{metadata.space}{default_type}"
            return value.dump(space=metadata.space, version=metadata.version, type=default_type)
        return str(value)

    def as_container_reference(self) -> ContainerProperty:
        if self.container is None or self.container_property is None:
            raise NeatValueError("Accessing container reference without container and container property set")
        return ContainerProperty(container=self.container, property_=self.container_property)

    def as_view_reference(self) -> ViewProperty:
        return ViewProperty(view=self.view, property_=self.view_property)


class PhysicalContainer(SheetRow):
    container: ContainerEntityType = Field(
        alias="Container", description="Container id, strongly advised to PascalCase usage."
    )
    name: str | None = Field(
        alias="Name", default=None, description="Human readable name of the container being defined."
    )
    description: str | None = Field(
        alias="Description", default=None, description="Short description of the node being defined."
    )
    constraint: ContainerConstraintListType | None = Field(
        None, alias="Constraint", description="List of required (comma separated) constraints for the container"
    )
    used_for: Literal["node", "edge", "all"] | None = Field(
        "all", alias="Used For", description=" Whether the container is used for nodes, edges or all."
    )

    @field_validator("constraint", mode="after")
    @classmethod
    def constraint_set_correctly(cls, value: ContainerConstraintListType | None) -> Any:
        if value is None:
            return value

        for constraint in value:
            if constraint.prefix is Undefined:
                message = f"The type of constraint is not defined. Please set 'requires:{constraint!s}'."
                warnings.warn(
                    message,
                    stacklevel=2,
                )
            elif constraint.prefix != "requires":
                message = (
                    f"Unsupported constraint type on container as "
                    f"the whole '{constraint.prefix}'. Currently only 'requires' is supported."
                )
                raise ValueError(message) from None

            if len(constraint.suffix) > CONSTRAINT_ID_MAX_LENGTH:
                message = f"Constraint id '{constraint.suffix}' exceeds maximum length of {CONSTRAINT_ID_MAX_LENGTH}."
                raise ValueError(message) from None

            if constraint.require is None:
                message = (
                    f"Container constraint must have a container set. "
                    f"Please set 'requires:{constraint!s}(container=space:external_id)'."
                )
                raise ValueError(message) from None

        return value

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.container,)

    def as_container(self) -> dm.ContainerApply:
        container_id = self.container.as_id()
        constraints: dict[str, dm.Constraint] = {}
        for constraint in self.constraint or []:
            if constraint.require is None:
                continue
            requires = dm.RequiresConstraint(constraint.require.as_id())
            constraints[constraint.suffix] = requires

        return dm.ContainerApply(
            space=container_id.space,
            external_id=container_id.external_id,
            name=self.name or None,
            description=self.description,
            constraints=constraints or None,
            properties={},
            used_for=self.used_for,
        )

    @field_serializer("container", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if metadata := _metadata(info.context):
            if isinstance(value, PhysicalEntity):
                return value.dump(space=metadata.space, version=metadata.version)
            elif isinstance(value, ConceptualEntity):
                return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("constraint", when_used="unless-none")
    def remove_default_spaces(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, list) and (metadata := _metadata(info.context)):
            return ",".join(
                constraint.dump(space=metadata.space, version=metadata.version)
                if isinstance(constraint, PhysicalEntity)
                else str(constraint)
                for constraint in value
            )
        return ",".join(str(value) for value in value)


class PhysicalView(SheetRow):
    view: ViewEntityType = Field(alias="View", description="View id, strongly advised to PascalCase usage.")
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the view being defined.")
    description: str | None = Field(
        alias="Description", default=None, description="Short description of the view being defined "
    )
    implements: ViewEntityList | None = Field(
        None,
        alias="Implements",
        description="List of parent view ids (comma separated) which the view being defined implements.",
    )
    filter_: HasDataFilter | NodeTypeFilter | RawFilter | None = Field(
        None, alias="Filter", description="Explicitly define the filter for the view."
    )
    in_model: bool = Field(
        True,
        alias="In Model",
        description="Indicates whether the view being defined is a part of the data model.",
    )
    conceptual: URIRefType | None = Field(
        None,
        alias="Conceptual",
        description="Used to make connection between physical and conceptual data model level",
    )

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.view,)

    @field_serializer("view", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if (metadata := _metadata(info.context)) and isinstance(value, ConceptualEntity):
            return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("implements", when_used="unless-none")
    def remove_default_spaces(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, list) and (metadata := _metadata(info.context)):
            return ",".join(
                parent.dump(space=metadata.space, version=metadata.version)
                if isinstance(parent, PhysicalEntity)
                else str(parent)
                for parent in value
            )
        return ",".join(str(value) for value in value) if isinstance(value, list) else value

    def as_view(self) -> dm.ViewApply:
        view_id = self.view.as_id()
        implements = [parent.as_id() for parent in self.implements or []] or None

        return dm.ViewApply(
            space=view_id.space,
            external_id=view_id.external_id,
            version=view_id.version or _DEFAULT_VERSION,
            name=self.name or None,
            description=self.description,
            filter=None if self.filter_ is None else self.filter_.as_dms_filter(),
            implements=implements,
            properties={},
        )

    def as_view_reference(self) -> ViewRef:
        return ViewRef(view=self.view)


class PhysicalNodeType(SheetRow):
    node: DMSNodeEntity = Field(alias="Node", description="The type definition of the node.")
    usage: Literal["type", "collection"] = Field(
        alias="Usage", description="What the usage of the node is in the data model."
    )
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the node being defined.")
    description: str | None = Field(
        alias="Description", default=None, description="Short description of the node being defined."
    )

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.node,)

    def as_node(self) -> dm.NodeApply:
        if self.usage == "type":
            return dm.NodeApply(space=self.node.space, external_id=self.node.external_id)
        elif self.usage == "collection":
            raise NotImplementedError("Collection nodes are not supported yet")
        else:
            raise ValueError(f"Unknown usage {self.usage}")

    @field_serializer("node", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, PhysicalEntity) and (metadata := _metadata(info.context)):
            return value.dump(space=metadata.space, version=metadata.version)
        return str(value)


class PhysicalEnum(SheetRow):
    collection: ConceptEntityType = Field(alias="Collection", description="The collection this enum belongs to.")
    value: str = Field(alias="Value", description="The value of the enum.")
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the enum.")
    description: str | None = Field(alias="Description", default=None, description="Short description of the enum.")

    def _identifier(self) -> tuple[Hashable, ...]:
        return self.collection, self.value

    @field_serializer("collection", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, PhysicalEntity) and (metadata := _metadata(info.context)):
            return value.dump(space=metadata.space, version=metadata.version)
        return str(value)


class PhysicalDataModel(BaseVerifiedDataModel):
    metadata: PhysicalMetadata = Field(alias="Metadata", description="Contains information about the data model.")
    properties: SheetList[PhysicalProperty] = Field(
        alias="Properties", description="Contains the properties of the data model."
    )
    views: SheetList[PhysicalView] = Field(alias="Views", description="Contains the views of the data model.")
    containers: SheetList[PhysicalContainer] | None = Field(
        None,
        alias="Containers",
        description="Contains the definition containers that are the physical storage of the data model.",
    )
    enum: SheetList[PhysicalEnum] | None = Field(
        None, alias="Enum", description="Contains the definition of enum values."
    )
    nodes: SheetList[PhysicalNodeType] | None = Field(
        None, alias="Nodes", description="Contains the definition of the node types."
    )

    @model_validator(mode="after")
    def set_neat_id(self) -> "PhysicalDataModel":
        namespace = self.metadata.namespace

        for view in self.views:
            if not view.neatId:
                view.neatId = namespace[view.view.suffix]

        for property_ in self.properties:
            if not property_.neatId:
                property_.neatId = namespace[f"{property_.view.suffix}/{property_.view_property}"]

        return self

    def update_neat_id(self) -> None:
        """Update neat ids"""

        namespace = self.metadata.namespace

        for view in self.views:
            view.neatId = namespace[view.view.suffix]

        for property_ in self.properties:
            property_.neatId = namespace[f"{property_.view.suffix}/{property_.view_property}"]

    def sync_with_conceptual_data_model(self, conceptual_data_model: "ConceptualDataModel") -> None:
        # Sync at the metadata level
        if conceptual_data_model.metadata.physical == self.metadata.identifier:
            self.metadata.conceptual = conceptual_data_model.metadata.identifier
        else:
            # if models are not linked to start with, we skip
            return None

        conceptual_properties_by_neat_id = {prop.neatId: prop for prop in conceptual_data_model.properties}
        physical_properties_by_neat_id = {prop.neatId: prop for prop in self.properties}
        for neat_id, prop in conceptual_properties_by_neat_id.items():
            if prop.physical in physical_properties_by_neat_id:
                physical_properties_by_neat_id[prop.physical].conceptual = neat_id

        classes_by_neat_id = {cls.neatId: cls for cls in conceptual_data_model.concepts}
        views_by_neat_id = {view.neatId: view for view in self.views}
        for neat_id, class_ in classes_by_neat_id.items():
            if class_.physical in views_by_neat_id:
                views_by_neat_id[class_.physical].conceptual = neat_id

    def as_schema(self, instance_space: str | None = None, remove_cdf_spaces: bool = False) -> DMSSchema:
        from ._exporter import _DMSExporter

        return _DMSExporter(self, instance_space, remove_cdf_spaces=remove_cdf_spaces).to_schema()

    @classmethod
    def display_type_name(cls) -> str:
        return "VerifiedPhysicalModel"

    def _repr_html_(self) -> str:
        summary = {
            "level": self.metadata.level,
            "intended for": "Data Engineer",
            "name": self.metadata.name,
            "space": self.metadata.space,
            "external_id": self.metadata.external_id,
            "version": self.metadata.version,
            "views": len(self.views),
            "containers": len(self.containers) if self.containers else 0,
            "properties": len(self.properties),
        }

        return pd.DataFrame([summary]).T.rename(columns={0: ""})._repr_html_()  # type: ignore



from __future__ import annotations

from collections.abc import Sequence

from cognite.neat._data_model.models.dms import ViewReference, ViewRequest, ViewResponse
from cognite.neat._utils.http_client import HTTPClient, SuccessResponse

from .api import Endpoint, NeatAPI
from .config import NeatClientConfig
from .data_classes import PagedResponse
from .filters import ViewFilter


class ViewsAPI(NeatAPI):
    def __init__(self, neat_config: NeatClientConfig, http_client: HTTPClient) -> None:
        super().__init__(
            neat_config,
            http_client,
            endpoint_map={
                "apply": Endpoint("POST", "/models/views", item_limit=100),
                "retrieve": Endpoint("POST", "/models/views/byids", item_limit=100),
                "delete": Endpoint("POST", "/models/views/delete", item_limit=100),
                "list": Endpoint("GET", "/models/views", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse) -> PagedResponse[ViewResponse]:
        return PagedResponse[ViewResponse].model_validate_json(response.body)

    def _validate_id_response(self, response: SuccessResponse) -> list[ViewReference]:
        return PagedResponse[ViewReference].model_validate_json(response.body).items

    def apply(self, items: Sequence[ViewRequest]) -> list[ViewResponse]:
        """Create or update views in CDF Project.

        Args:
            items: List of ViewRequest objects to create or update.
        Returns:
            List of ViewResponse objects.
        """
        return self._request_item_response(items, "apply")

    def retrieve(self, items: list[ViewReference], include_inherited_properties: bool = True) -> list[ViewResponse]:
        """Retrieve views by their identifiers.

        Args:
            items: List of (space, external_id, version) tuples identifying the views to retrieve.
            include_inherited_properties: If True, include properties inherited from parent views.

        Returns:
            List of ViewResponse objects.
        """
        return self._request_item_response(
            items, "retrieve", extra_body={"includeInheritedProperties": include_inherited_properties}
        )

    def delete(self, items: list[ViewReference]) -> list[ViewReference]:
        """Delete views by their identifiers.

        Args:
            items: List of (space, external_id, version) tuples identifying the views to delete.

        Returns:
            List of ViewReference objects representing the deleted views.
        """
        return self._request_id_response(items, "delete")

    def list(
        self,
        space: str | None = None,
        all_versions: bool = False,
        include_inherited_properties: bool = True,
        include_global: bool = False,
        limit: int | None = 10,
    ) -> list[ViewResponse]:
        """List views in CDF Project.

        Args:
            space: If specified, only views in this space are returned.
            all_versions: If True, return all versions. If False, only return the latest version.
            include_inherited_properties: If True, include properties inherited from parent views.
            include_global: If True, include global views.
            limit: Maximum number of views to return. If None, return all views.

        Returns:
            List of ViewResponse objects.
        """
        filter = ViewFilter(
            space=space,
            all_versions=all_versions,
            include_inherited_properties=include_inherited_properties,
            include_global=include_global,
        )
        return self._list(limit=limit, params=filter.dump())



from __future__ import annotations

from collections.abc import Sequence

from cognite.neat._data_model.models.dms import SpaceRequest, SpaceResponse
from cognite.neat._data_model.models.dms._references import SpaceReference
from cognite.neat._utils.http_client import HTTPClient, SuccessResponse

from .api import Endpoint, NeatAPI
from .config import NeatClientConfig
from .data_classes import PagedResponse
from .filters import DataModelingFilter


class SpacesAPI(NeatAPI):
    def __init__(self, neat_config: NeatClientConfig, http_client: HTTPClient) -> None:
        super().__init__(
            neat_config,
            http_client,
            endpoint_map={
                "apply": Endpoint("POST", "/models/spaces", item_limit=100),
                "retrieve": Endpoint("POST", "/models/spaces/byids", item_limit=100),
                "delete": Endpoint("POST", "/models/spaces/delete", item_limit=100),
                "list": Endpoint("GET", "/models/spaces", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse) -> PagedResponse[SpaceResponse]:
        return PagedResponse[SpaceResponse].model_validate_json(response.body)

    def _validate_id_response(self, response: SuccessResponse) -> list[SpaceReference]:
        return PagedResponse[SpaceReference].model_validate_json(response.body).items

    def apply(self, spaces: Sequence[SpaceRequest]) -> list[SpaceResponse]:
        """Apply (create or update) spaces in CDF.

        Args:
            spaces: List of SpaceRequest objects to apply.
        Returns:
            List of SpaceResponse objects.
        """
        return self._request_item_response(spaces, "apply")

    def retrieve(self, spaces: list[SpaceReference]) -> list[SpaceResponse]:
        """Retrieve spaces by their identifiers.

        Args:
            spaces: List of space identifiers to retrieve.

        Returns:
            List of SpaceResponse objects.
        """
        return self._request_item_response(spaces, "retrieve")

    def delete(self, spaces: list[SpaceReference]) -> list[SpaceReference]:
        """Delete spaces by their identifiers.

        Args:
            spaces: List of space identifiers to delete.
        Returns:
            List of SpaceReference objects representing the deleted spaces.
        """
        return self._request_id_response(spaces, "delete")

    def list(
        self,
        include_global: bool = False,
        limit: int | None = 10,
    ) -> list[SpaceResponse]:
        """List spaces in CDF Project.

        Args:
            include_global: If True, include global spaces.
            limit: Maximum number of spaces to return. If None, return all spaces.

        Returns:
            List of SpaceResponse objects.
        """
        filter = DataModelingFilter(include_global=include_global)
        return self._list(limit=limit, params=filter.dump())
